import hub
from abc import ABC
from typing import Any, List, Sequence, Optional
from hub.constants import ENCODING_DTYPE
import numpy as np


# the "last seen index" denotes an encoder row's last seen index
# this is the column that is binary searched over
LAST_SEEN_INDEX_COLUMN = -1


class Encoder(ABC):
    last_row = 0

    def is_index_in_last_row(self, arr, index) -> bool:
        """Checks if `index` is in the self.last_row of of encoder."""
        row = self.last_row
        return arr[row, -1] >= index and (row == 0 or arr[row - 1, -1] < index)

    def check_last_row(self, global_sample_index: int):
        """Takes a look at self.last_row and tries to find chunk id without binary search by looking at the current and next row."""
        arr = self.array
        if self.last_row < len(arr) and self.is_index_in_last_row(
            arr, global_sample_index
        ):
            return self.last_row
        elif self.last_row < len(arr) - 1:
            self.last_row += 1
            if self.is_index_in_last_row(arr, global_sample_index):
                return self.last_row
        return None

    def __init__(self, encoded=None, dtype=ENCODING_DTYPE):
        """Base class for custom encoders that allow reading meta information from sample indices without decoding the entire encoded state.

        Handles heavy lifting logic for:
            - Chunk ID encoder
            - Shape encoder
            - Byte positions encoder

        Lookup algorithm is essentially the same for all encoders, however the details are different.
        You can find all of this information in their respective classes.

        Layout:
            `_encoded` is a 2D array.

            Best case scenario:
                The best case scenario is when all samples have the same meta and can thus be stored in a single row,
                providing a O(1) lookup.

            Worst case scenario:
                The worst case scenario is when every sample has different meta values. This means the number of rows is equal to the number
                of samples, providing a O(log(N)) lookup.

            Lookup algorithm:
                To get the decoded meta for some sample index, you do a binary search over the column `LAST_SEEN_INDEX_COLUMN`. This will give you
                the row that corresponds to that sample index (since the right-most column is our "last index" for that meta information).
                Then, you decode the row and that is your meta!

        Raises:
            ValueError: If `encoded` is not the right dtype.

        Args:
            encoded (np.ndarray): Encoded state, if None state is empty. Helpful for deserialization. Defaults to None.
            dtype (np.dtype): Dtype of the encoder. Defaults to `ENCODING_DTYPE`.
        """
        self.dtype = dtype

        if isinstance(encoded, list):
            encoded = np.array(encoded, dtype=self.dtype)

        self._encoded = encoded
        if self._encoded is None:
            self._encoded = np.array([], dtype=self.dtype)

        if self._encoded.dtype != self.dtype:
            raise ValueError(
                f"Encoding dtype should be {self.dtype}, instead got {self._encoded.dtype}"
            )

        self.version = hub.__version__
        self.is_dirty = True

    @property
    def array(self):
        return self._encoded

    @property
    def nbytes(self):
        return self.array.nbytes

    @property
    def num_samples(self) -> int:
        if len(self._encoded) == 0:
            return 0
        return int(self._encoded[-1, LAST_SEEN_INDEX_COLUMN] + 1)

    def num_samples_at(self, row_index: int) -> int:
        """Calculates the number of samples a row in the encoding corresponds to.

        Args:
            row_index (int): This index will be used when indexing `self._encoded`.

        Returns:
            int: Representing the number of samples that a row's derivable value represents.
        """

        lower_bound = 0
        if len(self._encoded) > 1 and row_index > 0:
            lower_bound = self._encoded[row_index - 1, LAST_SEEN_INDEX_COLUMN] + 1
        upper_bound = self._encoded[row_index, LAST_SEEN_INDEX_COLUMN] + 1

        return int(upper_bound - lower_bound)

    def translate_index(self, local_sample_index: int) -> int:
        """Searches for the row index for where `local_sample_index` exists within `self._encoded`.
        This method is worst case log(N) due to the binary search.

        Args:
            local_sample_index (int): Index representing a sample. Localized to `self._encoded`.

        Raises:
            IndexError: Cannot index when there are no samples to index into.

        Returns:
            int: The index of the corresponding row inside the encoded state.
        """

        # TODO: optimize this (should accept an optional argument for starting point, instead of random binsearch)

        if len(self._encoded) == 0:
            raise IndexError(
                f"Index {local_sample_index} is out of bounds for an empty byte position encoding."
            )

        if local_sample_index < 0:
            local_sample_index += self.num_samples

        row_index = self.check_last_row(local_sample_index)
        if row_index is None:
            row_index = np.searchsorted(
                self._encoded[:, LAST_SEEN_INDEX_COLUMN], local_sample_index
            )
            self.last_row = row_index

        return row_index  # type: ignore

    def register_samples(self, item: Any, num_samples: int, row: Optional[int] = None):
        """Register `num_samples` as `item`. Combines when the `self._combine_condition` returns True.
        This method adds data to `self._encoded` without decoding.

        Args:
            item (Any): General input, will be passed along to subclass methods.
            num_samples (int): Number of samples that have `item`'s value. Will be passed along to subclass methods.
            row (Optional[int]): Parameter that shows to which chunk the samples need to be added
        """

        # TODO: optimize this

        self._validate_incoming_item(item, num_samples)

        if self.num_samples != 0:
            if self._combine_condition(item):
                last_index = self._encoded[-1, LAST_SEEN_INDEX_COLUMN]

                if row is not None:
                    self._encoded[row][1] += num_samples
                else:
                    new_last_index = self._derive_next_last_index(
                        last_index, num_samples
                    )
                    self._encoded[-1, LAST_SEEN_INDEX_COLUMN] = new_last_index
            else:
                decomposable = self._make_decomposable(item)

                last_index = self._encoded[-1, LAST_SEEN_INDEX_COLUMN]
                next_last_index = self._derive_next_last_index(last_index, num_samples)

                if row is not None:
                    self._encoded[:, LAST_SEEN_INDEX_COLUMN] += num_samples
                    shape_entry = np.array(
                        [*decomposable, num_samples - 1], dtype=self.dtype
                    )
                    self._encoded = np.insert(self._encoded, row, shape_entry, axis=0)
                else:
                    shape_entry = np.array(
                        [[*decomposable, next_last_index]], dtype=self.dtype
                    )
                    self._encoded = np.concatenate([self._encoded, shape_entry], axis=0)

        else:
            decomposable = self._make_decomposable(item)
            self._encoded = np.array(
                [[*decomposable, num_samples - 1]], dtype=self.dtype
            )

        self.is_dirty = True

    def _validate_incoming_item(self, item: Any, num_samples: int):
        """Raises appropriate exceptions for when `item` or `num_samples` are invalid.
        Subclasses should override this method when applicable.

        Args:
            item (Any): General input, will be passed along to subclass methods.
            num_samples (int): Number of samples that have `item`'s value. Will be passed along to subclass methods.

        Raises:
            ValueError: For the general case, `num_samples` should be > 0.
        """

        if num_samples <= 0:
            raise ValueError(f"`num_samples` should be > 0. Got: {num_samples}")

    def _combine_condition(self, item: Any, compare_row_index: int = -1) -> bool:
        """Should determine if `item` can be combined with a row in `self._encoded`."""

    def _derive_next_last_index(self, last_index, num_samples: int):
        """Calculates what the next last index should be."""
        return last_index + num_samples

    def _make_decomposable(self, item: Any, compare_row_index: int = -1) -> Sequence:
        """Should return a value that can be decompsed with the `*` operator. Example: `*(1, 2)`"""

        return item

    def _derive_value(
        self, row: np.ndarray, row_index: int, local_sample_index: int
    ) -> np.ndarray:
        """Given a row of `self._encoded`, this method should implement how `__getitem__` hands a value to the caller."""

    def __getitem__(
        self, local_sample_index: int, return_row_index: bool = False
    ) -> Any:
        """Derives the value at `local_sample_index`.

        Args:
            local_sample_index (int): Index of the sample for the desired value.
            return_row_index (bool): If True, the index of the row that the value was derived from is returned as well.
                Defaults to False.

        Returns:
            Any: Either just a singular derived value, or a tuple with the derived value and the row index respectively.
        """
        row_index = self.translate_index(local_sample_index)
        value = self._derive_value(
            self._encoded[row_index], row_index, local_sample_index
        )

        if return_row_index:
            return value, row_index

        return value

    def __setitem__(self, local_sample_index: int, item: Any):
        """Update the encoded value at a given index. Depending on the state, this may increase/decrease
        the size of the state.

        Updating:
            Updation is executed by going through a list of possible actions and trying to reduce the cost delta.

            Cost:
                Cost is defined as `len(self._encoded)`.
                The "cost delta" is the number of rows added/removed from `self._encoded` as a result of the action.

            Actions are chosen assuming `self._encoded` is already encoded properly.

            Note:
                An action that is "upwards" is being performed towards idx=0
                An action that is "downwards" is being performed away from idx=0

            Actions in order of execution:
                no change    (cost delta = 0)
                squeeze      (cost delta = -2)
                squeeze up   (cost delta = -1)
                squeeze down (cost delta = -1)
                move up      (cost delta = 0)
                move down    (cost delta = 0)
                replace      (cost delta = 0)
                split up     (cost delta = +1)
                split down   (cost delta = +1)
                split middle (cost delta = +2)

        Args:
            local_sample_index (int): Index representing a sample. Localized to `self._encoded`.
            item (Any): Item compatible with the encoder subclass.

        Raises:
            ValueError: If no update actions were taken.
        """
        row_index = self.translate_index(local_sample_index)
        # TODO: optimize this (vectorize __setitem__ to accept `Index` objects)
        actions = (
            self._try_not_changing,
            self._setup_update,  # not an actual action
            self._try_squeezing,
            self._try_squeezing_up,
            self._try_squeezing_down,
            self._try_moving_up,
            self._try_moving_down,
            self._try_replacing,
            self._try_splitting_up,
            self._try_splitting_down,
            self._try_splitting_middle,
        )

        action_taken = None
        for action in actions:
            if action(item, row_index, local_sample_index):  # type: ignore
                # each action returns a bool, if True that means the action was taken.
                action_taken = action
                break

        if action_taken is None:
            raise ValueError(
                f"Update could not be executed for idx={local_sample_index}, item={str(item)}"
            )

        self._post_process_state(start_row_index=max(row_index - 2, 0))
        self._reset_update_state()
        self.is_dirty = True

    def _post_process_state(self, start_row_index: int):
        """Overridden when more complex columns exist in subclasses. Example: byte positions."""

        pass

    def _reset_update_state(self):
        self._has_above = None
        self._has_below = None
        self._can_combine_above = None
        self._can_combine_below = None
        self._decomposable_item = None
        self._num_samples_at_row = None

    def _setup_update(self, item: Any, row_index: int, local_sample_index: int):
        """Setup the state variables for preceeding actions. Used for updating."""

        self._has_above = row_index > 0
        self._has_below = row_index + 1 < len(self._encoded)

        self._can_combine_above = (
            self._has_above
            and self._encoded[row_index - 1][LAST_SEEN_INDEX_COLUMN] + 1
            == local_sample_index
            and self._combine_condition(item, row_index - 1)
        )

        self._can_combine_below = (
            self._has_below
            and self._encoded[row_index][LAST_SEEN_INDEX_COLUMN] == local_sample_index
            and self._combine_condition(item, row_index + 1)
        )

        self._decomposable_item = self._make_decomposable(item, row_index)
        self._num_samples_at_row = self.num_samples_at(row_index)

    def _try_not_changing(self, item: Any, row_index: int, *_) -> bool:
        """If `item` already is the value at `row_index`, no need to make any updates.

        Cost delta = 0

        Example:
            Start:
                item    last index
                ------------------
                A       10
                B       11

            Update:
                self[5] = A

            End:
                item    last index
                ------------------
                A       10
                B       11
        """

        return self._combine_condition(item, row_index)

    def _try_squeezing(self, item: Any, row_index: int, *_) -> bool:
        """If update results in the above and below rows in `self._encoded`
        to match the incoming item, just combine them all into a single row.

        Cost delta = -2

        Example:
            Start:
                item    last index
                ------------------
                A       10
                B       11
                A       15

            Update:
                self[11] = A

            End:
                item    last index
                ------------------
                A       15
        """

        if self._num_samples_at_row != 1:
            return False

        if not (self._has_above and self._has_below):
            return False

        if not (self._can_combine_above and self._can_combine_below):
            return False

        # row can be "squeezed away"
        start = self._encoded[: row_index - 1]
        end = self._encoded[row_index + 1 :]
        self._encoded = np.concatenate((start, end))

        return True

    def _try_squeezing_up(self, item: Any, row_index: int, *_) -> bool:
        """If update results in the above row in `self._encoded`
        matching the incoming item, just combine them into a single row.

        Cost delta = -1

        Example:
            Start:
                item    last index
                ------------------
                A       10
                B       11
                C       15

            Update:
                self[11] = A

            End:
                item    last index
                ------------------
                A       11
                C       15
        """

        if self._num_samples_at_row != 1:
            return False

        if not self._has_above:
            return False

        if not self._can_combine_above:
            return False

        # row can be "squeezed upwards"
        start = self._encoded[:row_index]
        end = self._encoded[row_index + 1 :]
        start[-1, LAST_SEEN_INDEX_COLUMN] += 1
        self._encoded = np.concatenate((start, end))

        return True

    def _try_squeezing_down(self, item: Any, row_index: int, *_) -> bool:
        """If update results in the below row in `self._encoded`
        matching the incoming item, just combine them into a single row.

        Cost delta = -1

        Example:
            Start:
                item    last index
                ------------------
                A       10
                B       11
                C       15

            Update:
                self[11] = C

            End:
                item    last index
                ------------------
                A       10
                C       15
        """

        if self._num_samples_at_row != 1:
            return False

        if not self._has_below:
            return False

        if not self._can_combine_below:
            return False

        # row can be "squeezed downwards"
        start = self._encoded[:row_index]
        end = self._encoded[row_index + 1 :]
        self._encoded = np.concatenate((start, end))

        return True

    def _try_moving_up(self, item: Any, row_index: int, *_) -> bool:
        """If `item` exists in the row above `row_index`, then we can just use the above row to encode `item`.

        Cost delta = 0

        Example:
            Start:
                item    last index
                ------------------
                A       10
                B       15

            Update:
                self[11] = A

            End:
                item    last index
                ------------------
                A       11
                B       15
        """

        if self._can_combine_below or not self._can_combine_above:
            return False

        # sample can be "moved up"
        self._encoded[row_index - 1, LAST_SEEN_INDEX_COLUMN] += 1

        return True

    def _try_moving_down(self, item: Any, row_index: int, *_) -> bool:
        """If `item` exists in the row below `row_index`, then we can just use the below row to encode `item`.

        Cost delta = 0

        Example:
            Start:
                item    last index
                ------------------
                A       10
                B       15

            Update:
                self[10] = B

            End:
                item    last index
                ------------------
                A       9
                B       15
        """

        if self._can_combine_above or not self._can_combine_below:
            return False

        # sample can be "moved down"
        self._encoded[row_index, LAST_SEEN_INDEX_COLUMN] -= 1

        return True

    def _try_replacing(self, item: Any, row_index: int, *args) -> bool:
        """If the value encoded at `row_index` only exists for a single index, then `row_index`
        can be directly replaced with `item`.

        Cost delta = 0

        Example:
            Start:
                item    last index
                ------------------
                A       10
                B       11
                C       20

            Update:
                self[11] = D

            End:
                item    last index
                ------------------
                A       10
                D       11
                C       20
        """

        if self._num_samples_at_row != 1:
            return False

        # sample can be "replaced"
        self._encoded[row_index, :LAST_SEEN_INDEX_COLUMN] = item

        return True

    def _try_splitting_up(
        self, item: Any, row_index: int, local_sample_index: int
    ) -> bool:
        """If the row at `row_index` is being updated on the first index it is responsible for,
        AND the above row doesn't match `item`, a new row needs to be created above.

        Cost delta = +1

        Example:
            Start:
                item    last index
                ------------------
                A       10
                B       15
                C       20

            Update:
                self[11] = D

            End:
                item    last index
                ------------------
                A       10
                D       11
                B       15
                C       20
        """

        above_last_index = -1
        if self._has_above:
            above_last_index = self._encoded[row_index - 1, LAST_SEEN_INDEX_COLUMN]
        if above_last_index != local_sample_index - 1:
            return False

        # a new row should be created above
        start = self._encoded[: max(0, row_index)]
        end = self._encoded[row_index:]
        new_row = np.array(
            [*self._decomposable_item, local_sample_index], dtype=self.dtype
        )
        self._encoded = np.concatenate((start, [new_row], end))
        return True

    def _try_splitting_down(
        self, item: Any, row_index: int, local_sample_index: int
    ) -> bool:
        """If the row at `row_index` is being updated on the last index it is responsible for,
        AND the below row doesn't match `item`, a new row needs to be created below.

        Cost delta = +1

        Example:
            Start:
                item    last index
                ------------------
                A       10
                B       15
                C       20

            Update:
                self[15] = D

            End:
                item    last index
                ------------------
                A       10
                B       14
                D       15
                C       20
        """

        last_index = self._encoded[row_index, LAST_SEEN_INDEX_COLUMN]
        if last_index != local_sample_index:
            return False

        # a new row should be created below
        start = self._encoded[: row_index + 1]
        end = self._encoded[row_index + 1 :]
        start[-1, LAST_SEEN_INDEX_COLUMN] -= 1
        new_row = np.array(
            [*self._decomposable_item, local_sample_index], dtype=self.dtype
        )
        self._encoded = np.concatenate((start, [new_row], end))

        return True

    def _try_splitting_middle(
        self, item: Any, row_index: int, local_sample_index: int
    ) -> bool:
        """If the row at `row_index` is being updated on an index in the middle of the samples it is responsible for,
        a new row needs to be created above AND below.

        Cost delta = +2

        Example:
            Start:
                item    last index
                ------------------
                A       10
                B       15
                A       20

            Update:
                self[13] = A

            End:
                item    last index
                ------------------
                A       10
                B       12
                A       13
                B       15
                A       20
        """

        # 2 rows should be created, and 1 should be updated
        start = np.array(self._encoded[: row_index + 1])
        new_row = np.array(
            [*self._decomposable_item, local_sample_index], dtype=self.dtype
        )
        end = self._encoded[row_index:]
        start[-1, LAST_SEEN_INDEX_COLUMN] = local_sample_index - 1
        self._encoded = np.concatenate((start, [new_row], end))

        return True

    def _num_samples_in_last_row(self):
        if len(self._encoded) == 0:
            return 0
        elif len(self._encoded) == 1:
            return self._encoded[-1][LAST_SEEN_INDEX_COLUMN] + 1
        else:
            return (
                self._encoded[-1][LAST_SEEN_INDEX_COLUMN]
                - self._encoded[-2][LAST_SEEN_INDEX_COLUMN]
            )

    def pop(self, index: Optional[int] = None):
        if index is None:
            index = self.get_last_index_for_pop()
        _, row = self.__getitem__(index, return_row_index=True)  # type: ignore
        prev = -1 if row == 0 else self._encoded[row - 1, LAST_SEEN_INDEX_COLUMN]
        num_samples_in_row = self._encoded[row, LAST_SEEN_INDEX_COLUMN] - prev
        if num_samples_in_row == 0:
            raise ValueError("No samples to pop")
        self._encoded[row:, LAST_SEEN_INDEX_COLUMN] -= 1

        # after subtracting 1, the row is now empty
        if num_samples_in_row == 1:
            self._encoded = np.delete(self._encoded, row, axis=0)
        self.is_dirty = True

    def is_empty(self) -> bool:
        return len(self._encoded) == 0

    def tobytes(self) -> memoryview:
        raise NotImplementedError()

    @classmethod
    def frombuffer(cls, buffer: bytes):
        raise NotImplementedError()

    def get_last_index_for_pop(self) -> int:
        num_samples = self.num_samples
        if num_samples == 0:
            raise ValueError("No samples to pop")
        return num_samples - 1



from abc import abstractmethod
import struct
import numpy as np
from typing import List, Optional, Tuple, Union
import warnings

import hub
from hub.compression import (
    BYTE_COMPRESSION,
    IMAGE_COMPRESSION,
    VIDEO_COMPRESSION,
    get_compression_type,
)
from hub.constants import CONVERT_GRAYSCALE
from hub.core.fast_forwarding import ffw_chunk
from hub.core.linked_sample import LinkedSample
from hub.core.meta.encode.byte_positions import BytePositionsEncoder
from hub.core.meta.encode.shape import ShapeEncoder
from hub.core.meta.tensor_meta import TensorMeta
from hub.core.partial_reader import PartialReader
from hub.core.sample import Sample  # type: ignore
from hub.core.partial_sample import PartialSample
from hub.core.serialize import (
    deserialize_chunk,
    infer_chunk_num_bytes,
    infer_header_num_bytes,
    serialize_chunk,
    serialize_numpy_and_base_types,
    serialize_sample_object,
    serialize_text,
    serialize_tensor,
    serialize_partial_sample_object,
    get_header_from_url,
    serialize_text_sample_object,
)
from hub.core.storage.hub_memory_object import HubMemoryObject
from hub.core.tiling.sample_tiles import SampleTiles
from hub.util.exceptions import TensorInvalidSampleShapeError
from functools import reduce
from operator import mul

InputSample = Union[
    Sample,
    np.ndarray,
    int,
    float,
    bool,
    dict,
    list,
    str,
    np.integer,
    np.floating,
    np.bool_,
]
SerializedOutput = Tuple[bytes, Tuple]


class BaseChunk(HubMemoryObject):
    def __init__(
        self,
        min_chunk_size: int,
        max_chunk_size: int,
        tiling_threshold: int,
        tensor_meta: TensorMeta,
        compression: Optional[str] = None,
        encoded_shapes: Optional[np.ndarray] = None,
        encoded_byte_positions: Optional[np.ndarray] = None,
        data: Optional[Union[memoryview, PartialReader]] = None,
    ):
        super().__init__()
        self._data_bytes: Union[bytearray, bytes, memoryview, PartialReader] = (
            data or bytearray()
        )
        self.version = hub.__version__
        self.min_chunk_size = min_chunk_size
        self.max_chunk_size = max_chunk_size
        self.tiling_threshold = tiling_threshold

        self.tensor_meta = tensor_meta
        self.num_dims = len(tensor_meta.max_shape) if tensor_meta.max_shape else None
        self.is_text_like = (
            self.htype in {"json", "list", "text"} or self.tensor_meta.is_link
        )

        self.compression = compression
        compression_type = get_compression_type(compression)
        self.is_byte_compression = compression_type == BYTE_COMPRESSION
        self.is_image_compression = compression_type == IMAGE_COMPRESSION
        self.is_video_compression = compression_type == VIDEO_COMPRESSION
        self.is_convert_candidate = self.htype == "image" or self.is_image_compression

        self.shapes_encoder = ShapeEncoder(encoded_shapes)
        self.byte_positions_encoder = BytePositionsEncoder(encoded_byte_positions)

        if self.is_text_like and self.is_image_compression:
            raise ValueError("Can't use image compression with text data.")

        # These caches are only used for ChunkCompressed chunk.
        self.decompressed_samples: Optional[List[np.ndarray]] = None
        self.decompressed_bytes: Optional[bytes] = None

        # Whether tensor meta length is updated by chunk. Used by chunk engine while replacing chunks.
        self._update_tensor_meta_length: bool = (
            True  # Note: tensor meta shape interval is updated regardless.
        )
        self._item_size = None
        self._sample_size = None

    @property
    def is_fixed_shape(self):
        return (
            self.tensor_meta.min_shape == self.tensor_meta.max_shape
            and not self.is_text_like
        )

    @property
    def item_size(self):
        # should only be called if self.is_fixed_shape
        if self._item_size is None:
            if self.dtype is None:
                raise ValueError("Can't get item size as dtype is not set.")
            self._item_size = np.dtype(self.dtype).itemsize
        return self._item_size

    @property
    def sample_size(self):
        # should only be called if self.is_fixed_shape
        shape = self.tensor_meta.max_shape
        if self._sample_size is None:
            self._sample_size = self.item_size * reduce(mul, shape, 1)
        return self._sample_size

    def get_byte_positions(self, local_index):
        # should only be called if self.is_fixed_shape
        return local_index * self.sample_size, (local_index + 1) * self.sample_size

    @property
    def is_partially_read_chunk(self):
        return isinstance(self.data_bytes, PartialReader)

    @property
    def data_bytes(self) -> Union[bytearray, bytes, memoryview, PartialReader]:
        return self._data_bytes

    @data_bytes.setter
    def data_bytes(self, value: Union[bytearray, bytes, memoryview, PartialReader]):
        self._data_bytes = value

    @property
    def num_data_bytes(self) -> int:
        if isinstance(self.data_bytes, PartialReader):
            enc = self.byte_positions_encoder
            num_samples = enc.num_samples
            if num_samples == 0:
                return 0
            first_data_start_byte = enc[0][0]
            last_data_end_byte = enc[num_samples - 1][1]
            return last_data_end_byte - first_data_start_byte

        return len(self.data_bytes)

    @property
    def dtype(self):
        return self.tensor_meta.dtype

    @property
    def htype(self):
        return self.tensor_meta.htype

    @property
    def num_samples(self) -> int:
        if not self.shapes_encoder.is_empty():
            return self.shapes_encoder.num_samples
        else:
            return self.byte_positions_encoder.num_samples

    @property
    def nbytes(self):
        """Calculates the number of bytes `tobytes` will be without having to call `tobytes`. Used by `LRUCache` to determine if this chunk can be cached."""
        return infer_chunk_num_bytes(
            self.version,
            self.shapes_encoder.array,
            self.byte_positions_encoder.array,
            len_data=self.num_data_bytes,
        )

    @property
    def header_bytes(self):
        return infer_header_num_bytes(
            self.version, self.shapes_encoder.array, self.byte_positions_encoder.array
        )

    @property
    def memoryview_data(self):
        if isinstance(self.data_bytes, (memoryview, PartialReader)):
            return self.data_bytes
        return memoryview(self.data_bytes)

    @property
    def is_empty(self):
        return (
            self.num_data_bytes == 0
            and len(self.shapes_encoder.array) == 0
            and len(self.byte_positions_encoder.array) == 0
        )

    def tobytes(self) -> memoryview:
        if isinstance(self.data_bytes, PartialReader):
            self._make_data_bytearray()

        assert isinstance(self.data_bytes, (memoryview, bytearray, bytes))
        return serialize_chunk(
            self.version,
            self.shapes_encoder.array,
            self.byte_positions_encoder.array,
            [self.data_bytes],
        )

    @classmethod
    def frombuffer(cls, buffer: bytes, chunk_args: list, copy=True, url=False, partial=False):  # type: ignore
        if not buffer:
            return cls(*chunk_args)
        if url:
            version, shapes, byte_positions, header_size = get_header_from_url(
                buffer.decode("utf-8")
            )
            data = memoryview(buffer + struct.pack("<i", header_size))
        else:
            version, shapes, byte_positions, data = deserialize_chunk(buffer, copy=copy)
            if partial:
                data = None
        chunk = cls(*chunk_args, shapes, byte_positions, data=data)  # type: ignore
        chunk.version = version
        chunk.is_dirty = False
        return chunk

    @abstractmethod
    def extend_if_has_space(
        self, incoming_samples, update_meta: bool = True, end: bool = True
    ) -> float:
        """Extends the chunk with the incoming samples."""

    @abstractmethod
    def read_sample(
        self,
        local_index: int,
        cast: bool = True,
        copy: bool = False,
        decompress: bool = True,
        is_tile: bool = False,
    ):
        """Reads a sample from the chunk."""

    @abstractmethod
    def update_sample(self, local_index: int, new_sample: InputSample):
        """Updates a sample in the chunk."""

    def _make_data_bytearray(self):
        """Copies `self.data_bytes` into a bytearray if it is a memoryview."""
        # data_bytes will be a memoryview if frombuffer is called.
        if isinstance(self.data_bytes, PartialReader):
            chunk_bytes = self.data_bytes.get_all_bytes()
            self.data_bytes = bytearray(chunk_bytes[self.header_bytes :])
        elif isinstance(self.data_bytes, memoryview):
            self.data_bytes = bytearray(self.data_bytes)

    def prepare_for_write(self):
        ffw_chunk(self)
        self._make_data_bytearray()
        self.is_dirty = True

    def serialize_sample(
        self,
        incoming_sample: InputSample,
        sample_compression: Optional[str] = None,
        chunk_compression: Optional[str] = None,
        break_into_tiles: bool = True,
        store_uncompressed_tiles: bool = False,
    ) -> SerializedOutput:
        """Converts the sample into bytes"""
        dt, ht, min_chunk_size, tiling_threshold = (
            self.dtype,
            self.htype,
            self.min_chunk_size,
            self.tiling_threshold,
        )
        if tiling_threshold < 0:
            break_into_tiles = False

        if isinstance(incoming_sample, LinkedSample):
            if self.tensor_meta.is_link:
                incoming_sample = incoming_sample.path
            else:
                raise ValueError(
                    "hub.link() samples can only be appended to linked tensors. To create linked tensors, include link in htype during create_tensor, for example 'link[image]'."
                )

        if self.is_text_like:
            if isinstance(incoming_sample, LinkedSample):
                incoming_sample = incoming_sample.path
            if incoming_sample is None:
                htype = "text" if self.tensor_meta.is_link else self.htype
                empty_mapping = {"text": "", "list": [], "json": {}}
                incoming_sample = empty_mapping[htype]

            if isinstance(incoming_sample, Sample):
                if incoming_sample.is_text_like:
                    incoming_sample, shape = serialize_text_sample_object(  # type: ignore
                        incoming_sample, sample_compression
                    )
                else:
                    htype = "Linked" if self.tensor_meta.is_link else self.htype
                    raise TypeError(
                        f"Cannot append to {htype} tensor with Sample object"
                    )
            else:
                incoming_sample, shape = serialize_text(
                    incoming_sample, sample_compression, dt, ht  # type: ignore
                )
        elif incoming_sample is None:
            shape = (0,) * self.num_dims if self.num_dims else None
            incoming_sample = b""
        elif isinstance(incoming_sample, Sample):
            incoming_sample, shape = serialize_sample_object(  # type: ignore
                incoming_sample,
                sample_compression,
                chunk_compression,
                dt,
                ht,
                tiling_threshold,
                break_into_tiles,
                store_uncompressed_tiles,
            )
            shape = self.convert_to_rgb(shape)
        elif isinstance(incoming_sample, PartialSample):
            incoming_sample, shape = serialize_partial_sample_object(
                incoming_sample,
                sample_compression,
                chunk_compression,
                dt,
                ht,
                min_chunk_size,
            )
        elif isinstance(incoming_sample, hub.core.tensor.Tensor):
            incoming_sample, shape = serialize_tensor(
                incoming_sample,
                sample_compression,
                chunk_compression,
                dt,
                ht,
                tiling_threshold,
                break_into_tiles,
                store_uncompressed_tiles,
            )
        elif isinstance(
            incoming_sample,
            (np.ndarray, list, int, float, bool, np.integer, np.floating, np.bool_),
        ):
            incoming_sample, shape = serialize_numpy_and_base_types(
                incoming_sample,
                sample_compression,
                chunk_compression,
                dt,
                ht,
                tiling_threshold,
                break_into_tiles,
                store_uncompressed_tiles,
            )
        elif isinstance(incoming_sample, SampleTiles):
            shape = incoming_sample.sample_shape
        else:
            raise TypeError(f"Cannot serialize sample of type {type(incoming_sample)}")
        shape = self.normalize_shape(shape)
        return incoming_sample, shape  # type: ignore

    def convert_to_rgb(self, shape):
        if self.is_convert_candidate and CONVERT_GRAYSCALE:
            if self.num_dims is None:
                self.num_dims = len(shape)
            if len(shape) == 2 and self.num_dims == 3:
                message = "Grayscale images will be reshaped from (H, W) to (H, W, 1) to match tensor dimensions. This warning will be shown only once."
                warnings.warn(message)
                shape += (1,)  # type: ignore[assignment]
        return shape

    def can_fit_sample(self, sample_nbytes, buffer_nbytes=0):
        if self.num_data_bytes == 0:
            if self.tiling_threshold < 0:  # tiling disabled
                return True
            else:
                return buffer_nbytes + sample_nbytes <= self.tiling_threshold
        else:
            return (
                self.num_data_bytes + buffer_nbytes + sample_nbytes
                <= self.min_chunk_size
            )

    def copy(self, chunk_args=None):
        return self.frombuffer(self.tobytes(), chunk_args)

    def register_sample_to_headers(
        self,
        incoming_num_bytes: Optional[int],
        sample_shape: Tuple[int],
    ):
        """Registers a single sample to this chunk's header. A chunk should NOT exist without headers.

        Args:
            incoming_num_bytes (int): The length of the buffer that was used to
            sample_shape (Tuple[int]): Every sample that `num_samples` symbolizes is considered to have `sample_shape`.

        Raises:
            ValueError: If `incoming_num_bytes` is not divisible by `num_samples`.
        """
        # incoming_num_bytes is not applicable for image compressions
        if incoming_num_bytes is not None:
            self.byte_positions_encoder.register_samples(incoming_num_bytes, 1)
        if sample_shape is not None:
            if self.shapes_encoder.is_empty():
                num_samples = self.byte_positions_encoder.num_samples - 1
                self._fill_empty_shapes(sample_shape, num_samples)
            self.shapes_encoder.register_samples(sample_shape, 1)

    def register_in_meta_and_headers(
        self,
        sample_nbytes: Optional[int],
        shape,
        update_tensor_meta: bool = True,
    ):
        """Registers a new sample in meta and headers

        Args:
           sample_nbytes (Optional[int]): Paramter shat shows the numbero of bytes
           shape (Any): Parameter that shows the shape of the added elements
           update_commit_diff (bool): Parameter that shows if we need to update the commit diffs
           update_tensor_meta (bool): Parameter that shows if it is need to update tensor metas, in case of rechunk we do not need to update meta as we do not add new elements
        """
        self.register_sample_to_headers(sample_nbytes, shape)
        if update_tensor_meta:
            if self._update_tensor_meta_length:
                self.tensor_meta.update_length(1)
            if shape is not None:
                self.tensor_meta.update_shape_interval(shape)

    def update_in_meta_and_headers(
        self, local_index: int, sample_nbytes: Optional[int], shape
    ):
        """Updates an existing sample in meta and headers"""
        if sample_nbytes is not None:
            self.byte_positions_encoder[local_index] = sample_nbytes
        if shape is not None:
            if self.shapes_encoder.is_empty():
                num_samples = self.byte_positions_encoder.num_samples
                self._fill_empty_shapes(shape, num_samples)
            self.shapes_encoder[local_index] = shape
            self.tensor_meta.update_shape_interval(shape)

    def check_shape_for_update(self, shape):
        """Checks if the shape being assigned at the new index is valid."""
        if shape is None:
            return
        max_shape = self.tensor_meta.max_shape
        if max_shape:
            expected_dimensionality = len(max_shape)
            if expected_dimensionality != len(shape):
                raise TensorInvalidSampleShapeError(shape, expected_dimensionality)

    def create_updated_data(self, local_index: int, old_data, new_sample_bytes: bytes):
        if not old_data or self.byte_positions_encoder.is_empty():  # tiled sample
            return new_sample_bytes
        old_start_byte, old_end_byte = self.byte_positions_encoder[local_index]
        left_data = old_data[:old_start_byte]  # type: ignore
        right_data = old_data[old_end_byte:]  # type: ignore

        # preallocate
        total_new_bytes = len(left_data) + len(new_sample_bytes) + len(right_data)
        new_data = bytearray(total_new_bytes)

        # copy old data and add new data
        new_start_byte = old_start_byte
        new_end_byte = old_start_byte + len(new_sample_bytes)
        new_data[:new_start_byte] = left_data
        new_data[new_start_byte:new_end_byte] = new_sample_bytes
        new_data[new_end_byte:] = right_data
        return new_data

    def normalize_shape(self, shape):
        if shape is not None and len(shape) == 0:
            shape = (1,)
        return shape

    def write_tile(self, sample: SampleTiles):
        data, tile_shape = sample.yield_tile()
        self.data_bytes = data
        self.register_sample_to_headers(None, tile_shape)
        if sample.is_first_write:
            self.tensor_meta.update_shape_interval(sample.sample_shape)  # type: ignore
            if self._update_tensor_meta_length:
                self.tensor_meta.update_length(1)

    def pop_multiple(self, num_samples):
        self.prepare_for_write()

        if not self.byte_positions_encoder.is_empty():
            total_samples = self.shapes_encoder.num_samples
            starting_byte_first_popped_sample = self.byte_positions_encoder[
                total_samples - num_samples
            ][0]
            self.data_bytes = self.data_bytes[:starting_byte_first_popped_sample]

        for _ in range(num_samples):
            if not self.shapes_encoder.is_empty():
                self.shapes_encoder.pop()
            if not self.byte_positions_encoder.is_empty():
                self.byte_positions_encoder.pop()

    def _get_partial_sample_tile(self, as_bytes=False):
        if (
            not isinstance(self.data_bytes, PartialReader)
            and not self.data_bytes
            and len(self.shapes_encoder._encoded) > 0
        ):
            shape = self.shapes_encoder._encoded[0][:-1]
            if len(shape) and np.all(shape):
                if as_bytes:
                    return b"0" * int(
                        np.prod(np.array(shape, dtype=np.uint64))
                        * np.dtype(self.dtype).itemsize
                    )
                return np.zeros(shape, dtype=self.dtype)
        return None

    def pop(self, index):
        self.prepare_for_write()
        sb, eb = self.byte_positions_encoder[index]
        self.data_bytes = self.data_bytes[:sb] + self.data_bytes[eb:]
        if not self.shapes_encoder.is_empty():
            self.shapes_encoder.pop(index)
        if not self.byte_positions_encoder.is_empty():
            self.byte_positions_encoder.pop(index)

    def _fill_empty_shapes(self, shape, num_samples):
        dims = len(shape)
        self.num_dims = self.num_dims or dims
        if num_samples > 0:
            empty_shape = (0,) * dims
            self.shapes_encoder.register_samples(empty_shape, num_samples)
            self.tensor_meta.update_shape_interval(empty_shape)

    @property
    def is_empty_tensor(self):
        return len(self.tensor_meta.max_shape) == 0 and len(self.data_bytes) == 0



import pytest
import hub
import numpy as np
import uuid
from hub.tests.common import LinkTransformTestContext


def test_linking(memory_ds):
    ds = memory_ds
    with ds:
        ds.create_tensor("x")
        ds.create_tensor("y")
        ds._link_tensors("x", "y", "append_test")
        ds.x.extend(list(range(10)))
        np.testing.assert_array_equal(ds.x.numpy(), np.arange(10).reshape(-1, 1))
        np.testing.assert_array_equal(ds.x.numpy(), ds.y.numpy())


def test_linking_sequence(memory_ds):
    ds = memory_ds
    with ds:
        ds.create_tensor("x", htype="sequence")
        ds.create_tensor("x_id")
        id_f = lambda _: 0
        with LinkTransformTestContext(id_f, "id"):
            ds._link_tensors("x", "x_id", "id", flatten_sequence=False)
            ds.x.extend(np.random.random((10, 5, 3, 2)))
            assert len(ds.x) == len(ds.x_id) == 10
            np.testing.assert_array_equal(ds.x_id.numpy(), np.zeros((10, 1)))


def test_linking_sequence_update(memory_ds):
    ds = memory_ds
    with ds:
        ds.create_tensor("x", htype="sequence")
        ds.create_tensor("x_id")
        id_f = lambda _: 0
        id_f2 = lambda *_: 1  # updated samples will have x_id=1
        with LinkTransformTestContext(id_f, "id"):
            with LinkTransformTestContext(id_f2, "id2"):
                ds._link_tensors(
                    "x", "x_id", append_f="id", update_f="id2", flatten_sequence=False
                )
                ds.x.extend(np.random.random((10, 5, 3, 2)))
                ds.x[0] += 1
                ds.x[3] += 1
                expected = np.zeros((10, 1))
                expected[0] = 1
                expected[3] = 1
                np.testing.assert_array_equal(ds.x_id.numpy(), expected)



from hub.util.cache_chain import get_cache_chain
from hub.constants import MIN_FIRST_CACHE_SIZE, MIN_SECOND_CACHE_SIZE
import pytest

_MEMORY = "memory_storage"
_LOCAL = "local_storage"
_S3 = "s3_storage"
_HUB_CLOUD = "hub_cloud_storage"

CACHE_CHAINS = [
    (_MEMORY, _LOCAL),
    (_MEMORY, _S3),
    (_MEMORY, _HUB_CLOUD),
    (_LOCAL, _S3),
    (_LOCAL, _HUB_CLOUD),
    (_MEMORY, _LOCAL, _S3),
    (_MEMORY, _LOCAL, _HUB_CLOUD),
]
CACHE_CHAINS = list(map(lambda i: ",".join(i), CACHE_CHAINS))  # type: ignore


enabled_cache_chains = pytest.mark.parametrize(
    "cache_chain",
    CACHE_CHAINS,
    indirect=True,
)


@pytest.fixture
def cache_chain(request):
    requested_storages = request.param.split(",")

    storages = []
    cache_sizes = []

    # will automatically skip if a storage is not enabled
    for requested_storage in requested_storages:
        storage = request.getfixturevalue(requested_storage)
        storages.append(storage)

        if len(cache_sizes) == 0:
            cache_sizes.append(MIN_FIRST_CACHE_SIZE)
        elif len(cache_sizes) == 1:
            cache_sizes.append(MIN_SECOND_CACHE_SIZE)

    if len(storages) == len(cache_sizes):
        cache_sizes.pop()

    assert (
        len(storages) == len(cache_sizes) + 1
    ), f"Invalid test composition. {len(storages)} != {len(cache_sizes)} - 1"

    return get_cache_chain(storages, cache_sizes)



from typing import Any, Dict
import hub
from hub.core.storage.hub_memory_object import HubMemoryObject


class Meta(HubMemoryObject):
    def __init__(self):
        """Contains **required** key/values that datasets/tensors use to function.
        See the `Info` class for optional key/values for datasets/tensors.
        """
        super().__init__()
        self.version = hub.__version__

    def __getstate__(self) -> Dict[str, Any]:
        return {"version": self.version}



import hashlib


def hash_inputs(*args) -> str:
    sha = hashlib.sha3_256()
    for item in args:
        sha.update(str(item).encode("utf-8"))
    return sha.hexdigest()


def hash_str_to_int32(string: str):
    hash_ = int(hashlib.sha256(string.encode("utf-8")).hexdigest(), 16) >> 224
    return hash_



import os
import time
import hub
from hub.constants import TIMESTAMP_FILENAME
from hub.util.exceptions import DatasetHandlerError
from hub.util.storage import get_local_storage_path


def check_access_method(access_method: str, overwrite: bool):
    if access_method not in ["stream", "download", "local"]:
        raise ValueError(
            f"Invalid access method: {access_method}. Must be one of 'stream', 'download', 'local'"
        )
    if access_method in {"download", "local"}:
        if not os.environ.get("HUB_DOWNLOAD_PATH"):
            raise ValueError(
                f"HUB_DOWNLOAD_PATH environment variable is not set. Cannot use access method '{access_method}'"
            )
        if overwrite:
            raise ValueError(
                "Cannot use access methods download or local with overwrite=True as these methods only interact with local copy of the dataset."
            )


def parse_access_method(access_method: str):
    num_workers = None
    scheduler = None
    if access_method.startswith("download"):
        split = access_method.split(":")
        if len(split) == 1:
            split.extend(("threaded", "0"))
        elif len(split) == 2:
            split.append("threaded" if split[1].isnumeric() else "0")
        elif len(split) >= 3:
            num_integers = sum(1 for i in split if i.isnumeric())
            if num_integers != 1 or len(split) > 3:
                raise ValueError(
                    "Invalid access_method format. Expected format is one of the following: {download, download:scheduler, download:num_workers, download:scheduler:num_workers, download:num_workers:scheduler}"
                )

        access_method = "download"
        num_worker_index = 1 if split[1].isnumeric() else 2
        scheduler_index = 3 - num_worker_index
        num_workers = int(split[num_worker_index])
        scheduler = split[scheduler_index]
    return access_method, num_workers, scheduler


def get_local_dataset(
    access_method,
    path,
    read_only,
    memory_cache_size,
    local_cache_size,
    creds,
    token,
    verbose,
    ds_exists,
    num_workers,
    scheduler,
):
    local_path = get_local_storage_path(path, os.environ["HUB_DOWNLOAD_PATH"])
    if access_method == "download":
        if not ds_exists:
            raise DatasetHandlerError(
                f"Dataset {path} does not exist. Cannot use access method 'download'"
            )
        elif hub.exists(local_path):
            raise DatasetHandlerError(
                f"A dataset already exists at the download location {local_path}. To reuse the dataset, use access method 'local'. If you want to download the dataset again, delete the dataset at the download location and try again."
            )
        hub.deepcopy(
            path,
            local_path,
            src_creds=creds,
            src_token=token,
            num_workers=num_workers,
            scheduler=scheduler,
            progressbar=True,
            verbose=False,
        )
    elif not hub.exists(local_path):
        raise DatasetHandlerError(
            f"A dataset does not exist at the download location {local_path}. Cannot use access method 'local'. Use access method 'download' to first download the dataset and then use access method 'local' in subsequent runs."
        )

    ds = hub.load(
        local_path,
        read_only=read_only,
        verbose=verbose,
        memory_cache_size=memory_cache_size,
        local_cache_size=local_cache_size,
    )
    if access_method == "download":
        ds.storage.next_storage[TIMESTAMP_FILENAME] = time.ctime().encode("utf-8")
    else:
        timestamp = ds.storage.next_storage[TIMESTAMP_FILENAME].decode("utf-8")
        print(f"** Loaded local copy of dataset. Downloaded on: {timestamp}")
    return ds



import numpy as np
import hub
from hub.htype import HTYPE_CONFIGURATIONS
from typing import Any, List, Sequence, Tuple, Optional, Union


class ExternalCommandError(Exception):
    def __init__(self, command: str, status: int):
        super().__init__(
            f'Status for command "{command}" was "{status}", expected to be "0".'
        )


class KaggleError(Exception):
    message: str = ""


class KaggleMissingCredentialsError(KaggleError):
    def __init__(self, env_var_name: str):
        super().__init__(
            "Could not find %s in environment variables. Try setting them or providing the `credentials` argument. More information on how to get kaggle credentials: https://www.kaggle.com/docs/api"
            % env_var_name
        )


class KaggleDatasetAlreadyDownloadedError(KaggleError):
    def __init__(self, tag: str, path: str):
        self.message = f"Kaggle dataset {tag} already exists at {path}. You can get rid of this error by setting the `exist_ok` parameter to `True`."
        super().__init__(self.message)


class InvalidPathException(Exception):
    def __init__(self, directory):
        super().__init__(
            f"Dataset's path is an invalid path. It should be a valid local directory got {directory}."
        )


class AutoCompressionError(Exception):
    def __init__(self, directory):
        super().__init__(
            f"Auto compression could not run on {directory}. The directory doesn't contain any files."
        )


class InvalidFileExtension(Exception):
    def __init__(self, directory):
        super().__init__(
            f"Missing file with extension in {directory}. Expected a valid file extension got None."
        )


class SamePathException(Exception):
    def __init__(self, directory):
        super().__init__(
            f"Dataset source and destination path are same '{directory}'. Source and destination cannot be same for dataset ingestion, try setting different paths."
        )


class TensorInvalidSampleShapeError(Exception):
    def __init__(self, shape: Sequence[int], expected_dims: int):
        super().__init__(
            f"Sample shape length is expected to be {expected_dims}, actual length is {len(shape)}. Full incoming shape: {shape}"
        )


class TensorMetaMissingKey(Exception):
    def __init__(self, key: str, meta: dict):
        super().__init__(f"Key '{key}' missing from tensor meta '{str(meta)}'.")


class TensorDoesNotExistError(KeyError, AttributeError):
    def __init__(self, tensor_name: str):
        super().__init__(f"Tensor '{tensor_name}' does not exist.")


class TensorAlreadyExistsError(Exception):
    def __init__(self, key: str):
        super().__init__(
            f"Tensor '{key}' already exists. If applicable, you can use the `overwrite=True` parameter!"
        )


class TensorGroupDoesNotExistError(KeyError):
    def __init__(self, group_name: str):
        super().__init__(f"Tensor group '{group_name}' does not exist.")


class TensorGroupAlreadyExistsError(Exception):
    def __init__(self, key: str):
        super().__init__(
            f"Tensor group '{key}' already exists. A tensor group is created when a tensor has a '/' in its name, or using 'ds.create_group'."
        )


class InvalidTensorNameError(Exception):
    def __init__(self, name: str):
        if name:
            msg = (
                f"The use of a reserved attribute '{name}' as a tensor name is invalid."
            )
        else:
            msg = f"Tensor name cannot be empty."
        super().__init__(msg)


class InvalidTensorGroupNameError(Exception):
    def __init__(self, name: str):
        if name:
            msg = f"The use of a reserved attribute '{name}' as a tensor group name is invalid."
        else:
            msg = f"Tensor group name cannot be empty."
        super().__init__(msg)


class DynamicTensorNumpyError(Exception):
    def __init__(self, key: str, index, property_key: str):
        super().__init__(
            f"Tensor '{key}' with index = {str(index)} has dynamic '{property_key}' and cannot be converted into a `np.ndarray`. Try setting the parameter `aslist=True`"
        )


class InvalidShapeIntervalError(Exception):
    def __init__(
        self, message: str, lower: Sequence[int] = None, upper: Sequence[int] = None
    ):
        s = message

        if lower is not None:
            s += f" lower={str(lower)}"

        if upper is not None:
            s += f" upper={str(upper)}"

        super().__init__(s)


class InvalidKeyTypeError(TypeError):
    def __init__(self, item: Any):
        super().__init__(
            f"Item '{str(item)}' of type '{type(item).__name__}' is not a valid key."
        )


class UnsupportedTensorTypeError(TypeError):
    def __init__(self, item: Any):
        super().__init__(
            f"Key of type '{type(item).__name__}' is not currently supported to convert to a tensor."
        )


class InvalidBytesRequestedError(Exception):
    def __init__(self):
        super().__init__(
            "The byte range provided is invalid. Ensure that start_byte <= end_byte and start_byte > 0 and end_byte > 0"
        )


class ProviderListEmptyError(Exception):
    def __init__(self):
        super().__init__(
            "The provider_list passed to get_cache_chain needs to have 1 or more elements."
        )


class DirectoryAtPathException(Exception):
    def __init__(self):
        super().__init__(
            "The provided path is incorrect for this operation, there is a directory at the path. Provide a path to a file."
        )


class FileAtPathException(Exception):
    def __init__(self, path):
        super().__init__(
            f"Expected a directory at path {path} but found a file instead."
        )


class ProviderSizeListMismatch(Exception):
    def __init__(self):
        super().__init__("Ensure that len(size_list) + 1 == len(provider_list)")


class ModuleNotInstalledException(Exception):
    def __init__(self, message):
        super().__init__(message)


class LoginException(Exception):
    def __init__(
        self,
        message="Error while logging in, invalid auth token. Please try logging in again.",
    ):
        super().__init__(message)


class UserNotLoggedInException(Exception):
    def __init__(self, message=""):
        super().__init__(message)


class InvalidHubPathException(Exception):
    def __init__(self, path):
        super().__init__(
            f"The Dataset's path is an invalid Hub path. It should be of the form hub://username/dataset got {path}."
        )


class PathNotEmptyException(Exception):
    def __init__(self, use_hub=True):
        if use_hub:
            super().__init__(
                f"Please use a url that points to an existing Hub Dataset or an empty folder. If you wish to delete the folder and its contents, you may run hub.delete(dataset_path, force=True)."
            )
        else:
            super().__init__(
                f"Specified path is not empty. If you wish to delete the folder and its contents, you may run hub.delete(path, force=True)."
            )


# Exceptions encountered while interection with the Hub backend
class AuthenticationException(Exception):
    def __init__(self, message="Authentication failed. Please try logging in again."):
        super().__init__(message)


class AuthorizationException(Exception):
    def __init__(
        self,
        message="You are not authorized to access this resource on Activeloop Server.",
        response=None,
    ):
        self.response = response
        super().__init__(message)


class InvalidPasswordException(AuthorizationException):
    def __init__(
        self,
        message="The password you provided was invalid.",
    ):
        super().__init__(message)


class CouldNotCreateNewDatasetException(AuthorizationException):
    def __init__(
        self,
        path: str,
    ):

        extra = ""
        if path.startswith("hub://"):
            extra = "Since the path is a `hub://` dataset, if you believe you should have write permissions, try running `activeloop login`."

        message = f"Dataset at '{path}' doesn't exist, and you have no permissions to create one there. Maybe a typo? {extra}"
        super().__init__(message)


class ResourceNotFoundException(Exception):
    def __init__(
        self,
        message="The resource you are looking for was not found. Check if the name or id is correct.",
    ):
        super().__init__(message)


class BadRequestException(Exception):
    def __init__(self, message):
        message = (
            f"Invalid Request. One or more request parameters is incorrect.\n{message}"
        )
        super().__init__(message)


class OverLimitException(Exception):
    def __init__(
        self,
        message="You are over the allowed limits for this operation.",
    ):
        super().__init__(message)


class ServerException(Exception):
    def __init__(self, message="Internal Activeloop server error."):
        super().__init__(message)


class BadGatewayException(Exception):
    def __init__(self, message="Invalid response from Activeloop server."):
        super().__init__(message)


class GatewayTimeoutException(Exception):
    def __init__(self, message="Activeloop server took too long to respond."):
        super().__init__(message)


class WaitTimeoutException(Exception):
    def __init__(self, message="Timeout waiting for server state update."):
        super().__init__(message)


class LockedException(Exception):
    def __init__(self, message="The resource is currently locked."):
        super().__init__(message)


class UnexpectedStatusCodeException(Exception):
    def __init__(self, message):
        super().__init__(message)


class EmptyTokenException(Exception):
    def __init__(self, message="The authentication token is empty."):
        super().__init__(message)


# TODO Better S3 Exception handling


class S3Error(Exception):
    """Catchall for all errors encountered while working with S3"""


class S3GetError(S3Error):
    """Catchall for all errors encountered while working getting an object from S3"""


class S3SetError(S3Error):
    """Catchall for all errors encountered while working setting an object in S3"""


class S3DeletionError(S3Error):
    """Catchall for all errors encountered while working deleting an object in S3"""


class S3ListError(S3Error):
    """Catchall for all errors encountered while retrieving a list of objects present in S3"""


class CompressionError(Exception):
    pass


class UnsupportedCompressionError(CompressionError):
    def __init__(self, compression: Optional[str], htype: Optional[str] = None):
        if htype:
            super().__init__(
                f"Compression '{compression}' is not supported for {htype} htype."
            )
        else:
            super().__init__(
                f"Compression '{compression}' is not supported. Supported compressions: {hub.compressions}."
            )


class SampleCompressionError(CompressionError):
    def __init__(
        self,
        sample_shape: Tuple[int, ...],
        compression_format: Optional[str],
        message: str,
    ):
        super().__init__(
            f"Could not compress a sample with shape {str(sample_shape)} into '{compression_format}'. Raw error output: '{message}'.",
        )


class SampleDecompressionError(CompressionError):
    def __init__(self):
        super().__init__(
            f"Could not decompress sample buffer into an array. Either the sample's buffer is corrupted, or it is in an unsupported format. Supported compressions: {hub.compressions}."
        )


class InvalidImageDimensions(Exception):
    def __init__(self, actual_dims, expected_dims):
        super().__init__(
            f"The shape length {actual_dims} of the given array should "
            f"be greater than the number of expected dimensions {expected_dims}"
        )


class TensorUnsupportedSampleType(Exception):
    def __init__(self) -> None:
        super().__init__(
            f"Unable to append sample. Please specify numpy array, sequence of numpy arrays"
            "or resulting dictionary from .read() to be added to the tensor"
        )


class MetaError(Exception):
    pass


class MetaDoesNotExistError(MetaError):
    def __init__(self, key: str):
        super().__init__(
            f"A meta (key={key}) cannot be instantiated without `required_meta` when it does not exist yet. \
            If you are trying to read the meta, heads up: it didn't get written."
        )


class MetaAlreadyExistsError(MetaError):
    def __init__(self, key: str, required_meta: dict):
        super().__init__(
            f"A meta (key={key}) cannot be instantiated with `required_meta` when it already exists. \
            If you are trying to write the meta, heads up: it already got written (required_meta={required_meta})."
        )


class MetaInvalidKey(MetaError):
    def __init__(self, name: str, available_keys: List[str]):
        super().__init__(
            f'"{name}" is an invalid key for meta (`meta_object.{name}`). \
            Maybe a typo? Available keys: {str(available_keys)}'
        )


class MetaInvalidRequiredMetaKey(MetaError):
    def __init__(self, key: str, subclass_name: str):
        super().__init__(
            f"'{key}' should not be passed in `required_meta` (it is probably automatically set). \
            This means the '{subclass_name}' class was constructed improperly."
        )


class TensorMetaInvalidHtype(MetaError):
    def __init__(self, htype: str, available_htypes: Sequence[str]):
        super().__init__(
            f"Htype '{htype}' does not exist. Available htypes: {str(available_htypes)}"
        )


class TensorMetaInvalidHtypeOverwriteValue(MetaError):
    def __init__(self, key: str, value: Any, explanation: str = ""):
        super().__init__(
            f"Invalid value '{value}' for tensor meta key '{key}'. {explanation}"
        )


class TensorMetaMissingRequiredValue(MetaError):
    def __init__(self, htype: str, key: str):
        extra = ""
        if key == "sample_compression":
            extra = f"`sample_compression` may be `None` if you want your '{htype}' data to be uncompressed. Available compressors: {hub.compressions}"

        super().__init__(
            f"Htype '{htype}' requires you to specify '{key}' inside the `create_tensor` method call. {extra}"
        )


class TensorMetaInvalidHtypeOverwriteKey(MetaError):
    def __init__(self, htype: str, key: str, available_keys: Sequence[str]):
        super().__init__(
            f"Htype '{htype}' doesn't have a key for '{key}'. Available keys: {str(available_keys)}"
        )


class TensorDtypeMismatchError(MetaError):
    def __init__(self, expected: Union[np.dtype, str], actual: str, htype: str):
        msg = f"Dtype was expected to be '{expected}' instead it was '{actual}'. If you called `create_tensor` explicitly with `dtype`, your samples should also be of that dtype."

        # TODO: we may want to raise this error at the API level to determine if the user explicitly overwrote the `dtype` or not. (to make this error message more precise)
        # TODO: because if the user uses `dtype=np.uint8`, but the `htype` the tensor is created with has it's default dtype set as `uint8` also, then this message is ambiguous
        htype_dtype = HTYPE_CONFIGURATIONS[htype].get("dtype", None)
        if htype_dtype is not None and htype_dtype == expected:
            msg += f" Htype '{htype}' expects samples to have dtype='{htype_dtype}'."
            super().__init__("")

        super().__init__(msg)


class InvalidTensorLinkError(MetaError):
    def __init__(self, msg="Invalid tensor link."):
        super().__init__(msg)


class TensorMetaMutuallyExclusiveKeysError(MetaError):
    def __init__(
        self, keys: Optional[List[str]] = None, custom_message: Optional[str] = None
    ):
        if custom_message:
            msg = custom_message
        else:
            msg = f"Following fields are mutually exclusive: {keys}. "
        super().__init__(msg)


class ReadOnlyModeError(Exception):
    def __init__(self, custom_message: str = None):
        if custom_message is None:
            custom_message = "Modification when in read-only mode is not supported!"
        super().__init__(custom_message)


class TransformError(Exception):
    pass


class FilterError(Exception):
    pass


class InvalidInputDataError(TransformError):
    def __init__(self, operation):
        super().__init__(
            f"The data_in to transform is invalid. It doesn't support {operation} operation. "
            "Please use a list, a hub dataset or an object that supports both __getitem__ and __len__. "
            "Generators are not supported."
        )


class UnsupportedSchedulerError(TransformError):
    def __init__(self, scheduler):
        super().__init__(
            f"Hub compute currently doesn't support {scheduler} scheduler."
        )


class TensorMismatchError(TransformError):
    def __init__(self, tensors, output_keys, skip_ok=False):
        if skip_ok:
            super().__init__(
                f"One or more tensors generated during hub compute don't exist in the target dataset. With skip_ok=True, you can skip certain tensors in the transform, however you need to ensure that all tensors generated exist in the dataset.\n "
                f"Tensors in target dataset: {tensors}\n Tensors in output sample: {output_keys}"
            )
        else:
            super().__init__(
                f"One or more of the outputs generated during transform contain different tensors than the ones present in the target dataset of transform.\n "
                f"Tensors in target dataset: {tensors}\n Tensors in output sample: {output_keys}. If you want to do this, pass skip_ok=True to the eval method."
            )


class InvalidOutputDatasetError(TransformError):
    def __init__(
        self, message="The output Dataset to transform should not be `read_only`."
    ):
        super().__init__(message)


class InvalidTransformDataset(TransformError):
    def __init__(
        self,
        message="The TransformDataset (2nd argument to transform function) of one of the functions is invalid. All the tensors should have equal length for it to be valid.",
    ):
        super().__init__(message)


class HubComposeEmptyListError(TransformError):
    def __init__(self, message="Cannot hub.compose an empty list."):
        super().__init__(message)


class HubComposeIncompatibleFunction(TransformError):
    def __init__(self, index: int):
        super().__init__(
            f"The element passed to hub.compose at index {index} is incompatible. Ensure that functions are all decorated with hub.compute decorator and instead of passing my_fn, use my_fn() or my_fn(arg1=5, arg2=3) in the list."
        )


class DatasetUnsupportedPytorch(Exception):
    def __init__(self, reason):
        super().__init__(
            f"The Dataset object passed to Pytorch is incompatible. Reason: {reason}"
        )


class CorruptedMetaError(Exception):
    pass


class ChunkEngineError(Exception):
    pass


class FullChunkError(ChunkEngineError):
    pass


class ChunkIdEncoderError(ChunkEngineError):
    pass


class ChunkSizeTooSmallError(ChunkEngineError):
    def __init__(
        self,
        message="If the size of the last chunk is given, it must be smaller than the requested chunk size.",
    ):
        super().__init__(message)


class DatasetHandlerError(Exception):
    def __init__(self, message):
        super().__init__(message)


class MemoryDatasetCanNotBePickledError(Exception):
    def __init__(self):
        super().__init__(
            "Dataset having MemoryProvider as underlying storage should not be pickled as data won't be saved."
        )


class CorruptedSampleError(Exception):
    def __init__(self, compression):
        super().__init__(f"Invalid {compression} file.")


class VersionControlError(Exception):
    pass


class MergeError(Exception):
    pass


class MergeNotSupportedError(MergeError):
    def __init__(self):
        super().__init__(
            "This dataset was either created before merge functionality was added or id tensor wasn't created for one or more tensors. Create a new dataset to use merge."
        )


class MergeMismatchError(MergeError):
    def __init__(self, tensor_name, mismatch_type, original_value, target_value):
        message = f"Unable to merge, tensor {tensor_name} has different {mismatch_type}. Current:{original_value}, Target: {target_value}"
        super().__init__(message)


class MergeConflictError(MergeError):
    def __init__(self, conflict_dict=None, message=""):
        if conflict_dict:
            tensor_names = [k for k, v in conflict_dict.items() if v]
            message = f"Unable to merge, tensors {tensor_names} have conflicts and conflict resolution argument was not provided. Use conflict_resolution='theirs' or conflict_resolution='ours' to resolve the conflict."
            super().__init__(message)
        else:
            super().__init__(message)


class CheckoutError(VersionControlError):
    pass


class CommitError(VersionControlError):
    pass


class EmptyCommitError(CommitError):
    pass


class TensorModifiedError(Exception):
    def __init__(self):
        super().__init__(
            "The target commit is not an ancestor of the current commit, modified can't be calculated."
        )


class GCSDefaultCredsNotFoundError(Exception):
    def __init__(self):
        super().__init__(
            "Unable to find default google application credentials at ~/.config/gcloud/application_default_credentials.json. "
            "Please make sure you initialized gcloud service earlier."
        )


class InvalidOperationError(Exception):
    def __init__(self, method: str, type: str):
        if method == "read_only":
            super().__init__("read_only property cannot be toggled for a dataset view.")
        else:
            super().__init__(f"{method} method cannot be called on a {type} view.")


class AgreementError(Exception):
    pass


class AgreementNotAcceptedError(AgreementError):
    def __init__(self, agreements=None):
        self.agreements = agreements
        super().__init__(
            "You did not accept the agreement. Make sure you type in the dataset name exactly as it appears."
        )


class NotLoggedInAgreementError(AgreementError):
    def __init__(self):
        super().__init__(
            "You are not logged in. Please log in to accept the agreement."
        )


class NotLoggedInError(AgreementError):
    def __init__(self, msg=None):
        msg = msg or (
            "This dataset includes an agreement that needs to be accepted before you can use it.\n"
            "You need to be signed in to accept this agreement.\n"
            "You can login using 'activeloop login' on the command line if you have an account or using 'activeloop register' if you don't have one."
        )

        super().__init__(msg)


class RenameError(Exception):
    def __init__(self, msg="Only name of the dataset can be different in new path."):
        super().__init__(msg)


class BufferError(Exception):
    pass


class InfoError(Exception):
    pass


class OutOfChunkCountError(Exception):
    pass


class OutOfSampleCountError(Exception):
    pass


class SampleHtypeMismatchError(Exception):
    def __init__(self, htype, sample_type):
        super().__init__(
            f"htype '{htype}' does not support samples of type {sample_type}."
        )


class EmptyTensorError(Exception):
    def __init__(self, message):
        super().__init__(message)


class DatasetViewSavingError(Exception):
    pass


class ManagedCredentialsNotFoundError(Exception):
    def __init__(self, org_id, creds_key):
        super().__init__(
            f"Unable to find managed credentials '{creds_key}' for organization {org_id}."
        )


class UnableToReadFromUrlError(Exception):
    def __init__(self, url, status_code):
        super().__init__(f"Unable to read from url {url}. Status code: {status_code}")


class InvalidTokenException(Exception):
    def __init__(self):
        super().__init__(
            "Token is invalid. Make sure the full token string is included and try again."
        )


class TokenPermissionError(Exception):
    def __init__(self):
        message = (
            "A dataset does not exist at the specified path, or you do not have "
            "sufficient permissions to load or create one. Please check the dataset "
            "path and make sure that you have sufficient permissions to the path."
        )
        super().__init__(message)


class SampleAppendingError(Exception):
    def __init__(self):
        message = """Cannot append sample because tensor(s) are not specified. Expected input to ds.append is a dictionary. To append samples, you need to either specify the tensors and append the samples as a dictionary, like: `ds.append({"image_tensor": sample, "label_tensor": sample})` or you need to call `append` method of the required tensor, like: `ds.image_tensor.append(sample)`"""
        super().__init__(message)



from hub.core.storage import S3Provider, GCSProvider, GDriveProvider, MemoryProvider
from hub.experimental.util import raise_indra_installation_error  # type: ignore
from hub.util.dataset import try_flushing  # type: ignore

try:
    from indra import api  # type: ignore

    INDRA_INSTALLED = True
except ImportError:
    INDRA_INSTALLED = False


def dataset_to_hub3(hub2_dataset):
    """Convert a hub 2.x dataset object to a hub 3.x dataset object."""
    raise_indra_installation_error(INDRA_INSTALLED)
    try_flushing(hub2_dataset)
    path: str = hub2_dataset.path
    if path.startswith("gdrive://"):
        raise ValueError("Gdrive datasets are not supported for hub3")
    elif path.startswith("mem://"):
        raise ValueError("In memory datasets are not supported for hub3")
    elif path.startswith("hub://"):
        token = hub2_dataset._token
        provider = hub2_dataset.storage.next_storage
        if isinstance(provider, S3Provider):
            hub3_dataset = api.dataset(
                path,
                origin_path=provider.root,
                token=token,
                aws_access_key_id=provider.aws_access_key_id,
                aws_secret_access_key=provider.aws_secret_access_key,
                aws_session_token=provider.aws_session_token,
                region_name=provider.aws_region,
                endpoint_url=provider.endpoint_url,
                expiration=str(provider.expiration),
            )
        else:
            raise ValueError("GCP datasets are not supported for hub3 currently.")
    elif path.startswith("s3://"):
        s3_provider = hub2_dataset.storage.next_storage
        aws_access_key_id = s3_provider.aws_access_key_id
        aws_secret_access_key = s3_provider.aws_secret_access_key
        aws_session_token = s3_provider.aws_session_token
        region_name = s3_provider.aws_region
        endpoint_url = s3_provider.endpoint_url

        # we don't need to pass profile name as hub has already found creds for it
        hub3_dataset = api.dataset(
            path,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            aws_session_token=aws_session_token,
            region_name=region_name,
            endpoint_url=endpoint_url,
        )

    elif path.startswith(("gcs://", "gs://", "gcp://")):
        raise ValueError("GCP datasets are not supported for hub3 currently.")
    else:
        hub3_dataset = api.dataset(path)

    commit_id = hub2_dataset.pending_commit_id
    hub3_dataset.checkout(commit_id)
    slice_ = hub2_dataset.index.values[0].value

    if slice_ != slice(None):
        if isinstance(slice_, tuple):
            slice_ = list(slice_)
        hub3_dataset = hub3_dataset[slice_]
    return hub3_dataset


def verify_base_storage(dataset):
    if isinstance(dataset.base_storage, (GCSProvider, GDriveProvider, MemoryProvider)):
        raise ValueError(
            "GCS, Google Drive and Memory datasets are not supported for experimental features currently."
        )



from typing import Any, List, Tuple, Optional
from hub.core.meta.encode.base_encoder import Encoder, LAST_SEEN_INDEX_COLUMN
from hub.constants import ENCODING_DTYPE
from hub.util.exceptions import (
    ChunkIdEncoderError,
    OutOfChunkCountError,
    OutOfSampleCountError,
)
from hub.core.storage.hub_memory_object import HubMemoryObject
import numpy as np
from hub.core.serialize import serialize_chunkids, deserialize_chunkids
from hub.util.generate_id import generate_id

CHUNK_ID_COLUMN = 0


class ChunkIdEncoder(Encoder, HubMemoryObject):
    def __init__(self, encoded=None, dtype=ENCODING_DTYPE):
        super().__init__(encoded, dtype)

    @staticmethod
    def name_from_id(id) -> str:
        """Returns the hex of `id` with the "0x" prefix removed. This is the chunk's name and should be used to determine the chunk's key.
        Can convert back into `id` using `id_from_name`. You can get the `id` for a chunk using `__getitem__`."""
        return hex(id)[2:]

    @staticmethod
    def id_from_name(name: str):
        """Returns the 64-bit integer from the hex `name` generated by `name_from_id`."""
        return int("0x" + name, 16)

    def get_name_for_chunk(self, chunk_index: int) -> str:
        """Gets the name for the chunk at index `chunk_index`. If you need to get the name for a chunk from a sample index, instead
        use `__getitem__`, then `name_from_id`."""

        chunk_id = self._encoded[:, CHUNK_ID_COLUMN][chunk_index]
        return ChunkIdEncoder.name_from_id(chunk_id)

    def get_id_for_chunk(self, chunk_index: int) -> str:
        """Gets the if for the chunk at index `chunk_index`. If you need to get the name for a chunk from a sample index, instead
        use `__getitem__`, then `name_from_id`."""

        return self._encoded[:, CHUNK_ID_COLUMN][chunk_index]

    @property
    def num_chunks(self) -> int:
        if self.num_samples == 0:
            return 0
        return len(self._encoded)

    def get_next_chunk_id(self, row) -> Optional[str]:
        if (
            self.num_chunks is None
            or self._encoded is None
            or not row < self.num_chunks - 1
        ):
            return None

        return self._encoded[row + 1][0]

    def get_prev_chunk_id(self, row) -> Optional[str]:
        if self.num_chunks is None or self._encoded is None or row == 0:
            return None

        return self._encoded[row - 1][0]

    def decrease_samples(self, row: int = 0, num_samples: int = 0):
        """Decrease sample count from encoder

        Args:
            row (int): row of the chunk
            num_samples (int): number of samples to be reduced

        Raises:
            OutOfSampleCountError: when num_samples are exeeding sample count
            OutOfChunkCountError: When the row is out of chunk bounds
        """
        if self.num_samples_at(row) < num_samples:
            raise OutOfSampleCountError()

        if self.num_chunks < row + 1:
            raise OutOfChunkCountError()

        self._encoded[row][LAST_SEEN_INDEX_COLUMN] -= num_samples

        self.is_dirty = True

    def delete_chunk_id(self, row):
        """Delete row from encoder

        Args:
            row (int): the row of chunk that needs to be deleted

        Raises:
            OutOfChunkCountError: When the row is out of chunk bounds
        """
        if row > self.num_chunks:
            raise OutOfChunkCountError

        self._encoded = np.delete(self._encoded, row, axis=0)
        self.is_dirty = True

    def generate_chunk_id(
        self, register: Optional[bool] = True, row: Optional[int] = None
    ):
        """Generates a random 64bit chunk ID using uuid4. Also prepares this ID to have samples registered to it.
        This method should be called once per chunk created.

        Args:
            register (Optional, bool): Whether the generated chunk id should be added to the encoder. Default True.
            row (Optional, int): Iterator position where the new generated id should be inserted

        Returns:
            The generated chunk id.

        Raises:
            OutOfChunkCountError: When the row is out of chunk bounds
        """
        chunk_id = generate_id(self.dtype)
        if register:
            if self.num_samples == 0:
                assert row is None
                self._encoded = np.array([[chunk_id, -1]], dtype=self.dtype)

            else:
                if row is not None and row < self.num_chunks:
                    new_entry = np.array(
                        [chunk_id, self._encoded[row][LAST_SEEN_INDEX_COLUMN]]
                    )
                    self._encoded = np.insert(self._encoded, row + 1, new_entry, axis=0)
                    return chunk_id
                if row is not None and row != self.num_chunks:
                    raise OutOfChunkCountError()
                last_index = self.num_samples - 1

                new_entry = np.array(
                    [[chunk_id, last_index]],
                    dtype=self.dtype,
                )
                self._encoded = np.concatenate([self._encoded, new_entry])
        return chunk_id

    def register_samples(self, num_samples: int, row: Optional[int] = None):  # type: ignore
        """Registers samples to the chunk ID that was generated last with the `generate_chunk_id` method.
        This method should be called at least once per chunk created.

        Args:
            num_samples (int): The number of samples the last chunk ID should have added to it's registration.
            row (int, Optional): The row of chunk to register the samples in.

        Raises:
            ValueError: `num_samples` should be non-negative.
            ChunkIdEncoderError: Must call `generate_chunk_id` before registering samples.
            ChunkIdEncoderError: `num_samples` can only be 0 if it is able to be a sample continuation accross chunks.
        """

        super().register_samples(None, num_samples, row=row)

    def translate_index_relative_to_chunks(self, global_sample_index: int) -> int:
        """Converts `global_sample_index` into a new index that is relative to the chunk the sample belongs to.

        Example:
            Given: 2 sampes in chunk 0, 2 samples in chunk 1, and 3 samples in chunk 2.
            >>> self.num_samples
            7
            >>> self.num_chunks
            3
            >>> self.translate_index_relative_to_chunks(0)
            0
            >>> self.translate_index_relative_to_chunks(1)
            1
            >>> self.translate_index_relative_to_chunks(2)
            0
            >>> self.translate_index_relative_to_chunks(3)
            1
            >>> self.translate_index_relative_to_chunks(6)
            2

        Args:
            global_sample_index (int): Index of the sample relative to the containing tensor.

        Returns:
            int: local index value between 0 and the amount of samples the chunk contains - 1.
        """

        ls = self.__getitem__(global_sample_index, return_row_index=True)  # type: ignore

        assert len(ls) == 1, len(
            ls
        )  # this method should only be called for non tiled samples
        chunk_index = ls[0][1]

        if chunk_index == 0:
            return global_sample_index

        current_entry = self._encoded[chunk_index - 1]  # type: ignore
        last_num_samples = current_entry[LAST_SEEN_INDEX_COLUMN] + 1

        return int(global_sample_index - last_num_samples)

    def _validate_incoming_item(self, _, num_samples: int):
        if num_samples < 0:
            raise ValueError(
                f"Cannot register negative num samples. Got: {num_samples}"
            )

        if len(self.array) == 0:
            raise ChunkIdEncoderError(
                "Cannot register samples because no chunk IDs exist."
            )

        if num_samples == 0 and self.num_chunks < 2:
            raise ChunkIdEncoderError(
                "Cannot register 0 num_samples (signifying a partial sample continuing the last chunk) when no last chunk exists."
            )

        # note: do not call super() method (num_samples can be 0)

    def _derive_next_last_index(self, last_index, num_samples: int):
        # this operation will trigger an overflow for the first addition, so supress the warning
        np.seterr(over="ignore")
        new_last_index = last_index + self.dtype(num_samples)
        np.seterr(over="warn")

        return new_last_index

    def _combine_condition(self, *args) -> bool:
        """Always returns True because sample registration can always be done. Used in base encoder `register_samples`."""

        return True

    def _derive_value(self, row: np.ndarray, *_) -> np.ndarray:
        return row[CHUNK_ID_COLUMN]

    def __setitem__(self, *args):
        raise NotImplementedError(
            "There is no reason for ChunkIdEncoder to be updated now."
        )

    def __getitem__(
        self, local_sample_index: int, return_row_index: bool = False
    ) -> Any:
        """Derives the value at `local_sample_index`.

        Args:
            local_sample_index (int): Index of the sample for the desired value.
            return_row_index (bool): If True, the index of the row that the value was derived from is returned as well.
                Defaults to False.

        Returns:
            Any: Either just a singular derived value, or a tuple with the derived value and the row index respectively.
        """
        if local_sample_index < 0:
            local_sample_index += self.num_samples

        row_index = self.translate_index(local_sample_index)
        output: List[Any] = []
        value = self._derive_value(
            self._encoded[row_index], row_index, local_sample_index
        )
        if return_row_index:
            output.append((value, row_index))
        else:
            output.append(value)

        row_index += 1

        while row_index < len(self._encoded):
            if self._encoded[row_index][LAST_SEEN_INDEX_COLUMN] == local_sample_index:
                self.last_row = row_index
                value = self._derive_value(
                    self._encoded[row_index], row_index, local_sample_index
                )
                if return_row_index:
                    output.append((value, row_index))
                else:
                    output.append(value)
                row_index += 1
            else:
                break
        return output

    def _num_samples_in_last_chunk(self):
        return self._num_samples_in_last_row()

    def pop(self, index: Optional[int] = None) -> Tuple[List, List, bool]:
        """Pops the last sample added to the encoder and returns ids of chunks to be deleted from storage.
        Returns:
            Tuple of list of affected chunk ids, their rows and boolean specifying whether those chunks should be deleted
        """
        if index is None:
            index = self.get_last_index_for_pop()
        out = self.__getitem__(index, return_row_index=True)  # type: ignore
        chunk_ids = [out[i][0] for i in range(len(out))]
        rows = [out[i][1] for i in range(len(out))]
        if len(chunk_ids) > 1:  # tiled sample
            self._encoded[rows[-1] + 1 :, LAST_SEEN_INDEX_COLUMN] -= 1
            self._encoded = np.delete(self._encoded, rows, axis=0)
            to_delete = True
        else:
            row = rows[0]
            prev = -1 if row == 0 else self._encoded[row - 1][LAST_SEEN_INDEX_COLUMN]
            num_samples_in_chunk = self.array[row][LAST_SEEN_INDEX_COLUMN] - prev

            if num_samples_in_chunk == 1:
                self._encoded = np.delete(self._encoded, row, axis=0)
                to_delete = True
            elif num_samples_in_chunk > 1:
                self._encoded[row:, LAST_SEEN_INDEX_COLUMN] -= 1
                to_delete = False
            else:
                raise ValueError("No samples to pop")

        self.is_dirty = True
        return chunk_ids, rows, to_delete

    def _replace_chunks_for_tiled_sample(
        self, global_sample_index: int, chunk_ids: List
    ):
        current_chunk_ids_and_rows = self.__getitem__(  # type: ignore
            global_sample_index, return_row_index=True
        )
        start_row = current_chunk_ids_and_rows[0][1]
        end_row = current_chunk_ids_and_rows[-1][1]
        if len(current_chunk_ids_and_rows) == len(chunk_ids):
            # inplace update
            self._encoded[start_row : end_row + 1, CHUNK_ID_COLUMN] = chunk_ids
        else:
            top = self._encoded[:start_row]
            bottom = self._encoded[end_row + 1 :]
            mid = np.empty((len(chunk_ids), 2), dtype=self.dtype)
            mid[:, CHUNK_ID_COLUMN] = chunk_ids
            mid[:, LAST_SEEN_INDEX_COLUMN] = global_sample_index
            self._encoded = np.concatenate([top, mid, bottom], axis=0)
        self.is_dirty = True

    @classmethod
    def frombuffer(cls, buffer: bytes):
        instance = cls()
        if not buffer:
            return instance
        version, ids, dtype = deserialize_chunkids(buffer)
        if ids.nbytes:
            instance._encoded = ids
        instance.version = version
        instance.is_dirty = False
        instance.dtype = dtype
        return instance

    def tobytes(self) -> memoryview:
        return serialize_chunkids(self.version, self._encoded)



import posixpath
from typing import Any, Dict, Union
from hub.client.utils import get_user_name
from hub.constants import HUB_CLOUD_DEV_USERNAME
from hub.core.dataset import Dataset
from hub.client.client import HubBackendClient
from hub.util.bugout_reporter import hub_reporter
from hub.util.exceptions import RenameError
from hub.util.link import save_link_creds
from hub.util.path import is_hub_cloud_path
from hub.util.tag import process_hub_path
from hub.util.logging import log_visualizer_link
from warnings import warn
import time
import hub


class HubCloudDataset(Dataset):
    """Subclass of :class:`Dataset`. Hub cloud datasets are those datasets which are stored on Activeloop servers, their paths look like:
    ``hub://username/dataset_name``."""

    def _first_load_init(self, verbose=True):
        self._set_org_and_name()
        if self.is_first_load:
            if self.is_actually_cloud:
                if self.verbose and verbose:
                    log_visualizer_link(self.path)
            else:
                # NOTE: this can happen if you override `hub.core.dataset.FORCE_CLASS`
                warn(
                    f'Created a hub cloud dataset @ "{self.path}" which does not have the "hub://" prefix. Note: this dataset should only be used for testing!'
                )
            self.link_creds.populate_all_managed_creds()

    @property
    def client(self):
        if self._client is None:
            self.__dict__["_client"] = HubBackendClient(token=self._token)
        return self._client

    @property
    def is_actually_cloud(self) -> bool:
        """Datasets that are connected to hub cloud can still technically be stored anywhere.
        If a dataset is hub cloud but stored without ``hub://`` prefix, it should only be used for testing.
        """
        return is_hub_cloud_path(self.path)  # type: ignore

    @property
    def token(self):
        """Get attached token of the dataset"""
        if self._token is None:
            self.__dict__["_token"] = self.client.get_token()
        return self._token

    def _set_org_and_name(self):
        if self.is_actually_cloud:
            if self.org_id is not None:
                return
            _, org_id, ds_name, subdir = process_hub_path(self.path)
            if subdir:
                ds_name += "/" + subdir
        else:
            # if this dataset isn't actually pointing to a datset in the cloud
            # a.k.a this dataset is trying to simulate a hub cloud dataset
            # it's safe to assume they want to use the dev org
            org_id = HUB_CLOUD_DEV_USERNAME
            ds_name = self.path.replace("/", "_").replace(".", "")
        self.__dict__["org_id"] = org_id
        self.__dict__["ds_name"] = ds_name

    def _is_sub_ds(self):
        return "/" in self.ds_name

    def _register_dataset(self):
        # called in super()._populate_meta
        self._set_org_and_name()
        if self._is_sub_ds():
            return
        self.client.create_dataset_entry(
            self.org_id,
            self.ds_name,
            self.version_state["meta"].__getstate__(),
            public=self.public,
        )
        self._send_dataset_creation_event()

    def _send_event(
        self,
        event_id: str,
        event_group: str,
        hub_meta: Dict[str, Any],
        has_head_changes: bool = None,
    ):
        username = get_user_name()
        has_head_changes = (
            has_head_changes if has_head_changes is not None else self.has_head_changes
        )
        common_meta = {
            "username": username,
            "commit_id": self.commit_id,
            "pending_commit_id": self.pending_commit_id,
            "has_head_changes": has_head_changes,
        }
        hub_meta.update(common_meta)
        event_dict = {
            "id": event_id,
            "event_group": event_group,
            "ts": time.time(),
            "hub_meta": hub_meta,
            "creator": "Hub",
        }
        hub.event_queue.put((self.client, event_dict))

    def _send_query_progress(
        self,
        query_id: str = "",
        query_text: str = "",
        start: bool = False,
        end: bool = False,
        progress: int = 0,
        status="",
    ):
        hub_meta = {
            "query_id": query_id,
            "query_text": query_text,
            "progress": progress,
            "start": start,
            "end": end,
            "status": status,
        }
        event_id = f"{self.org_id}/{self.ds_name}.query"
        self._send_event(event_id=event_id, event_group="query", hub_meta=hub_meta)

    def _send_compute_progress(
        self,
        compute_id: str = "",
        start: bool = False,
        end: bool = False,
        progress: int = 0,
        status="",
    ):
        hub_meta = {
            "compute_id": compute_id,
            "progress": progress,
            "start": start,
            "end": end,
            "status": status,
        }
        event_id = f"{self.org_id}/{self.ds_name}.compute"
        self._send_event(
            event_id=event_id, event_group="hub_compute", hub_meta=hub_meta
        )

    def _send_pytorch_progress(
        self,
        pytorch_id: str = "",
        start: bool = False,
        end: bool = False,
        progress: int = 0,
        status="",
    ):
        hub_meta = {
            "pytorch_id": pytorch_id,
            "progress": progress,
            "start": start,
            "end": end,
            "status": status,
        }
        event_id = f"{self.org_id}/{self.ds_name}.pytorch"
        self._send_event(event_id=event_id, event_group="pytorch", hub_meta=hub_meta)

    def _send_commit_event(self, commit_message: str, commit_time, author: str):
        # newly created commit can't have head_changes
        hub_meta = {
            "commit_message": commit_message,
            "commit_time": str(commit_time),
            "author": author,
        }
        event_id = f"{self.org_id}/{self.ds_name}.commit"
        self._send_event(
            event_id=event_id,
            event_group="dataset_commit",
            hub_meta=hub_meta,
            has_head_changes=False,
        )

    def _send_branch_creation_event(self, branch_name: str):
        hub_meta = {"branch_name": branch_name}
        event_id = f"{self.org_id}/{self.ds_name}.branch_created"
        self._send_event(
            event_id=event_id,
            event_group="dataset_branch_creation",
            hub_meta=hub_meta,
            has_head_changes=False,
        )

    def _send_dataset_creation_event(self):
        hub_meta = {}
        event_id = f"{self.org_id}/{self.ds_name}.dataset_created"
        self._send_event(
            event_id=event_id,
            event_group="dataset_creation",
            hub_meta=hub_meta,
            has_head_changes=False,
        )

    def make_public(self):
        self._set_org_and_name()
        if not self.public:
            self.client.update_privacy(self.org_id, self.ds_name, public=True)
            self.__dict__["public"] = True

    def make_private(self):
        self._set_org_and_name()
        if self.public:
            self.client.update_privacy(self.org_id, self.ds_name, public=False)
            self.__dict__["public"] = False

    def delete(self, large_ok=False):
        super().delete(large_ok=large_ok)
        if self._is_sub_ds():
            return
        self.client.delete_dataset_entry(self.org_id, self.ds_name)

    def rename(self, path):
        self.storage.check_readonly()
        path = path.rstrip("/")
        root, new_name = posixpath.split(path)
        if root != posixpath.split(self.path)[0]:
            raise RenameError
        self.client.rename_dataset_entry(self.org_id, self.ds_name, new_name)

        self.ds_name = new_name
        self.path = path

    def __getstate__(self) -> Dict[str, Any]:
        self._set_org_and_name()
        state = super().__getstate__()
        return state

    def __setstate__(self, state: Dict[str, Any]):
        super().__setstate__(state)
        self._client = None
        self._first_load_init(verbose=False)

    def visualize(
        self, width: Union[int, str, None] = None, height: Union[int, str, None] = None
    ):
        from hub.visualizer import visualize

        hub_reporter.feature_report(feature_name="visualize", parameters={})
        visualize(self.path, self.token, width=width, height=height)

    def add_creds_key(self, creds_key: str, managed: bool = False):
        """Adds a new creds key to the dataset. These keys are used for tensors that are linked to external data.

        Examples:

            >>> # create/load a dataset
            >>> ds = hub.dataset("hub://username/dataset")
            >>> # add a new creds key
            >>> ds.add_creds_key("my_s3_key")

        Args:
            creds_key (str): The key to be added.
            managed (bool): If ``True``, the creds corresponding to the key will be fetched from activeloop platform.
                Note, this is only applicable for datasets that are connected to activeloop platform.
                Defaults to ``False``.
        """
        self.link_creds.add_creds_key(creds_key, managed=managed)
        save_link_creds(self.link_creds, self.storage)
        self.link_creds.warn_missing_managed_creds()

    def update_creds_key(self, old_creds_key: str, new_creds_key: str):
        """Replaces the old creds key with the new creds key. This is used to replace the creds key used for external data."""
        if old_creds_key in self.link_creds.managed_creds_keys:
            raise ValueError(
                f"""Cannot update managed creds key directly. If you want to update it, follow these steps:-
                1. ds.change_creds_management("{old_creds_key}", False)
                2. ds.update_creds_key("{old_creds_key}", "{new_creds_key}")
                3. [OPTIONAL] ds.change_creds_management("{new_creds_key}", True)
                """
            )
        super().update_creds_key(old_creds_key, new_creds_key)
        self.link_creds.warn_missing_managed_creds()

    def change_creds_management(self, creds_key: str, managed: bool):
        """Changes the management status of the creds key.

        Args:
            creds_key (str): The key whose management status is to be changed.
            managed (bool): The target management status. If ``True``, the creds corresponding to the key will be fetched from activeloop platform.

        Raises:
            ValueError: If the dataset is not connected to activeloop platform.
            KeyError: If the creds key is not present in the dataset.

        Examples:

            >>> # create/load a dataset
            >>> ds = hub.dataset("hub://username/dataset")
            >>> # add a new creds key
            >>> ds.add_creds_key("my_s3_key")
            >>> # Populate the name added with creds dictionary
            >>> # These creds are only present temporarily and will have to be repopulated on every reload
            >>> ds.populate_creds("my_s3_key", {})
            >>> # Change the management status of the key to True. Before doing this, ensure that the creds have been created on activeloop platform
            >>> # Now, this key will no longer use the credentials populated in the previous step but will instead fetch them from activeloop platform
            >>> # These creds don't have to be populated again on every reload and will be fetched every time the dataset is loaded
            >>> ds.change_creds_management("my_s3_key", True)

        """

        key_index = self.link_creds.creds_mapping[creds_key] - 1
        changed = self.link_creds.change_creds_management(creds_key, managed)
        if changed:
            save_link_creds(
                self.link_creds, self.storage, managed_info=(managed, key_index)
            )

    def _load_link_creds(self):
        """Loads the link creds from the storage."""
        super()._load_link_creds()
        if self.link_creds.client is None:
            self._set_org_and_name()
            self.link_creds.org_id = self.org_id
            self.link_creds.client = self.client



import hub
from hub.core.fast_forwarding import ffw_tensor_meta
from typing import Any, Callable, Dict, List, Sequence, Union, Optional
import numpy as np
from hub.util.exceptions import (
    TensorMetaInvalidHtype,
    TensorMetaInvalidHtypeOverwriteValue,
    TensorMetaInvalidHtypeOverwriteKey,
    TensorMetaMissingRequiredValue,
    TensorMetaMutuallyExclusiveKeysError,
    UnsupportedCompressionError,
    TensorInvalidSampleShapeError,
    InvalidTensorLinkError,
)
from hub.util.json import validate_json_schema
from hub.constants import (
    REQUIRE_USER_SPECIFICATION,
    UNSPECIFIED,
)
from hub.compression import (
    COMPRESSION_ALIASES,
)
from hub.htype import HTYPE_CONFIGURATIONS, HTYPE_SUPPORTED_COMPRESSIONS, htype as HTYPE
from hub.htype import HTYPE_CONFIGURATIONS, REQUIRE_USER_SPECIFICATION, UNSPECIFIED
from hub.core.meta.meta import Meta
from hub.core.tensor_link import get_link_transform


class TensorMeta(Meta):
    name: Optional[str] = None
    htype: str
    dtype: str
    min_shape: List[int]
    max_shape: List[int]
    length: int
    sample_compression: str
    chunk_compression: str
    max_chunk_size: int
    tiling_threshold: int
    hidden: bool
    links: Dict[str, Dict[str, Union[str, bool]]]
    is_sequence: bool
    is_link: bool
    verify: bool

    def __init__(
        self,
        htype: str = UNSPECIFIED,
        **kwargs,
    ):
        """Tensor metadata is responsible for keeping track of global sample metadata within a tensor.

        Note:
            Tensor metadata that is automatically synchronized with `storage`. For more details, see the `Meta` class.
            Auto-populates `required_meta` that `Meta` accepts as an argument.

        Args:
            htype (str): All tensors require an `htype`. This determines the default meta keys/values.
            **kwargs: Any key that the provided `htype` has can be overridden via **kwargs. For more information, check out `hub.htype`.
        """

        super().__init__()
        if htype and htype not in [UNSPECIFIED, HTYPE.DEFAULT]:
            self.set_htype(htype, **kwargs)
        else:
            self.set_htype(HTYPE.DEFAULT, **kwargs)
            self.htype = None  # type: ignore

    def add_link(
        self, name, append_f: str, update_f: Optional[str], flatten_sequence: bool
    ):
        link = {
            "append": append_f,
            "flatten_sequence": flatten_sequence,
        }
        if update_f is not None:
            link["update"] = update_f
        d = {name: link}
        _validate_links(d)
        self.links.update(d)  # type: ignore
        self.is_dirty = True

    def set_hidden(self, val: bool):
        self.hidden = val
        self.is_dirty = True

    def set_dtype(self, dtype: np.dtype):
        """Should only be called once."""

        if self.dtype is not None:
            raise ValueError(
                f"Tensor meta already has a dtype ({self.dtype}). Incoming: {dtype.name}."
            )

        self.dtype = dtype.name
        self.is_dirty = True

    def set_dtype_str(self, dtype_name: str):
        self.dtype = dtype_name
        self.is_dirty = True

    def set_htype(self, htype: str, **kwargs):
        """Should only be called once."""

        if getattr(self, "htype", None) is not None:
            raise ValueError(
                f"Tensor meta already has a htype ({self.htype}). Incoming: {htype}."
            )

        if not kwargs:
            kwargs = HTYPE_CONFIGURATIONS[htype]

        _validate_htype_exists(htype)
        _validate_htype_overwrites(htype, kwargs)
        _replace_unspecified_values(htype, kwargs)
        _validate_required_htype_overwrites(htype, kwargs)
        _format_values(htype, kwargs)

        required_meta = _required_meta_from_htype(htype)
        required_meta.update(kwargs)

        self._required_meta_keys = tuple(required_meta.keys())

        for k in self._required_meta_keys:
            if getattr(self, k, None):
                required_meta.pop(k, None)

        self.__dict__.update(required_meta)
        self.is_dirty = True
        if self.links is None:
            self.links = {}
        _validate_links(self.links)

    def update_shape_interval(self, shape: Sequence[int]):
        initial_min_shape = None if self.min_shape is None else self.min_shape.copy()
        initial_max_shape = None if self.max_shape is None else self.max_shape.copy()

        if not self.min_shape:  # both min_shape and max_shape are set together
            self.min_shape = list(shape)
            self.max_shape = list(shape)
        else:
            expected_dims = len(self.min_shape)

            if len(shape) != expected_dims:
                raise TensorInvalidSampleShapeError(shape, len(self.min_shape))

            for i, dim in enumerate(shape):
                self.min_shape[i] = min(dim, self.min_shape[i])
                self.max_shape[i] = max(dim, self.max_shape[i])

        if initial_min_shape != self.min_shape or initial_max_shape != self.max_shape:
            self.is_dirty = True

    def update_length(self, length: int):
        self.length += length
        if length != 0:
            self.is_dirty = True

    def pop(self, index):
        self.length -= 1
        if self.length == 0:
            self.min_shape = []
            self.max_shape = []
        self.is_dirty = True

    def __getstate__(self) -> Dict[str, Any]:
        d = super().__getstate__()

        for key in self._required_meta_keys:
            d[key] = getattr(self, key)
        d["name"] = self.name

        return d

    def __setstate__(self, state: Dict[str, Any]):
        if "chunk_compression" not in state:
            state["chunk_compression"] = None  # Backward compatibility
        if "hidden" not in state:
            state["hidden"] = False
        super().__setstate__(state)
        self._required_meta_keys = tuple(state.keys())
        ffw_tensor_meta(self)

    @property
    def nbytes(self):
        # TODO: optimize this
        return len(self.tobytes())

    def __str__(self):
        return str(self.__getstate__())


def _validate_links(links: dict):
    if not isinstance(links, dict):
        raise InvalidTensorLinkError()
    allowed_keys = ("append", "update", "flatten_sequence")
    for out_tensor, args in links.items():
        if not isinstance(out_tensor, str):
            raise InvalidTensorLinkError()
        if not isinstance(args, dict):
            raise InvalidTensorLinkError()
        if "append" not in args:
            raise InvalidTensorLinkError(
                f"append transform not specified for link {out_tensor}"
            )
        if "flatten_sequence" not in args:
            raise InvalidTensorLinkError(
                f"flatten_sequence arg not specified for link {out_tensor}"
            )
        try:
            get_link_transform(args["append"])
        except KeyError:
            raise InvalidTensorLinkError(f"Invalid append transform: {args['append']}")
        if "update" in args:
            try:
                get_link_transform(args["update"])
            except KeyError:
                raise InvalidTensorLinkError(
                    f"Invalid update transform: {args['append']}"
                )
        for k in args:
            if k not in allowed_keys:
                raise InvalidTensorLinkError(f"Invalid key in link meta: {k}")


def _required_meta_from_htype(htype: str) -> dict:
    """Gets a dictionary with all required meta information to define a tensor."""

    _validate_htype_exists(htype)
    defaults = HTYPE_CONFIGURATIONS[htype]

    required_meta = {
        "htype": htype,
        "min_shape": [],
        "max_shape": [],
        "length": 0,
        "hidden": False,
        **defaults,
    }

    return required_meta


def _validate_htype_overwrites(htype: str, htype_overwrite: dict):
    """Raises errors if `htype_overwrite` has invalid keys or was missing required values."""

    defaults = HTYPE_CONFIGURATIONS[htype]

    for key, value in htype_overwrite.items():
        if key not in defaults:
            raise TensorMetaInvalidHtypeOverwriteKey(htype, key, list(defaults.keys()))

        if value == UNSPECIFIED:
            if defaults[key] == REQUIRE_USER_SPECIFICATION:
                raise TensorMetaMissingRequiredValue(htype, key)

    sc = htype_overwrite["sample_compression"]
    cc = htype_overwrite["chunk_compression"]
    compr = sc if cc in (None, UNSPECIFIED) else cc
    if htype.startswith("image") and sc == UNSPECIFIED and cc == UNSPECIFIED:
        if not htype_overwrite["is_link"]:
            raise TensorMetaMissingRequiredValue(
                htype, ["chunk_compression", "sample_compression"]  # type: ignore
            )
    if htype in ("audio", "video", "point_cloud"):
        if cc not in (UNSPECIFIED, None):
            raise UnsupportedCompressionError("Chunk compression", htype=htype)
        elif sc == UNSPECIFIED:
            if not htype_overwrite["is_link"]:
                raise TensorMetaMissingRequiredValue(
                    htype, "sample_compression"  # type: ignore
                )
    supported_compressions = HTYPE_SUPPORTED_COMPRESSIONS.get(htype)
    if (
        compr
        and compr != UNSPECIFIED
        and supported_compressions
        and compr not in supported_compressions
    ):
        raise UnsupportedCompressionError(compr, htype=htype)


def _replace_unspecified_values(htype: str, htype_overwrite: dict):
    """Replaces `UNSPECIFIED` values in `htype_overwrite` with the `htype`'s defaults."""

    defaults = HTYPE_CONFIGURATIONS[htype]

    for k, v in htype_overwrite.items():
        if v == UNSPECIFIED:
            htype_overwrite[k] = defaults[k]

    if (
        htype in ("json", "list", "text", "point_cloud_calibration_matrix")
        and not htype_overwrite["dtype"]
    ):
        htype_overwrite["dtype"] = HTYPE_CONFIGURATIONS[htype]["dtype"]


def _validate_required_htype_overwrites(htype: str, htype_overwrite: dict):
    """Raises errors if `htype_overwrite` has invalid values."""
    sample_compression = htype_overwrite["sample_compression"]
    sample_compression = COMPRESSION_ALIASES.get(sample_compression, sample_compression)
    if sample_compression not in hub.compressions:
        raise UnsupportedCompressionError(sample_compression)

    chunk_compression = htype_overwrite["chunk_compression"]
    chunk_compression = COMPRESSION_ALIASES.get(chunk_compression, chunk_compression)
    if chunk_compression not in hub.compressions:
        raise UnsupportedCompressionError(chunk_compression)

    if sample_compression and chunk_compression:
        raise TensorMetaMutuallyExclusiveKeysError(
            custom_message="Specifying both sample-wise and chunk-wise compressions for the same tensor is not yet supported."
        )

    if htype_overwrite["dtype"] is not None:
        if htype in ("json", "list"):
            validate_json_schema(htype_overwrite["dtype"])
        else:
            _raise_if_condition(
                "dtype",
                htype_overwrite,
                lambda dtype: not _is_dtype_supported_by_numpy(dtype),
                "Datatype must be supported by numpy. Can be an `str`, `np.dtype`, or normal python type (like `bool`, `float`, `int`, etc.). List of available numpy dtypes found here: https://numpy.org/doc/stable/user/basics.types.html",
            )

    if htype == "text":
        if htype_overwrite["dtype"] not in (str, "str"):
            raise TensorMetaInvalidHtypeOverwriteValue(
                "dtype",
                htype_overwrite["dtype"],
                "dtype for tensors with text htype should always be `str`",
            )


def _format_values(htype: str, htype_overwrite: dict):
    """Replaces values in `htype_overwrite` with consistent types/formats."""

    dtype = htype_overwrite["dtype"]
    if dtype is not None:
        if htype in ("json", "list", "point_cloud_calibration_matrix"):
            if getattr(dtype, "__module__", None) == "typing":
                htype_overwrite["dtype"] = str(dtype)
        else:
            htype_overwrite["dtype"] = np.dtype(htype_overwrite["dtype"]).name

    for key, value in COMPRESSION_ALIASES.items():
        if htype_overwrite.get("sample_compression") == key:
            htype_overwrite["sample_compression"] = value
        if htype_overwrite.get("chunk_compression") == key:
            htype_overwrite["chunk_compression"] = value


def _validate_htype_exists(htype: str):
    """Raises errors if given an unrecognized htype."""
    if htype not in HTYPE_CONFIGURATIONS:
        raise TensorMetaInvalidHtype(htype, list(HTYPE_CONFIGURATIONS.keys()))


def _raise_if_condition(
    key: str, meta: dict, condition: Callable[[Any], bool], explanation: str = ""
):
    v = meta[key]
    if condition(v):
        raise TensorMetaInvalidHtypeOverwriteValue(key, v, explanation)


def _is_dtype_supported_by_numpy(dtype: str) -> bool:
    try:
        np.dtype(dtype)
        return True
    except:
        return False



import json
from typing import Optional
import warnings
from hub.constants import ALL_CLOUD_PREFIXES
from hub.core.storage.hub_memory_object import HubMemoryObject
from hub.core.storage.provider import StorageProvider
from hub.core.storage.s3 import S3Provider
from hub.util.token import expires_in_to_expires_at, is_expired_token


class LinkCreds(HubMemoryObject):
    def __init__(self):
        self.creds_keys = []
        self.creds_dict = {}  # keys to actual creds dictionary
        self.creds_mapping = {}  # keys to numbers, for encoding
        self.managed_creds_keys = set()  # keys which are managed
        self.used_creds_keys = set()  # keys which are used by one or more samples
        self.storage_providers = {}
        self.default_s3_provider = None
        self.default_gcs_provider = None
        self.client = None
        self.org_id = None

    def get_creds(self, key: Optional[str]):
        if key in {"ENV", None}:
            return {}
        if key not in self.creds_keys:
            raise KeyError(f"Creds key {key} does not exist")
        if key not in self.creds_dict:
            raise ValueError(
                f"Creds key {key} hasn't been populated. Populate it using ds.populate_creds()"
            )
        if (
            self.client is not None
            and key in self.managed_creds_keys
            and is_expired_token(self.creds_dict[key])
        ):
            self.refresh_managed_creds(key)  # type: ignore
        return self.creds_dict[key]

    def refresh_managed_creds(self, creds_key: str):
        if creds_key not in self.managed_creds_keys:
            raise ValueError(f"Creds key {creds_key} is not managed")
        creds = self.fetch_managed_creds(creds_key)
        self.populate_creds(creds_key, creds)

    def get_default_provider(self, provider_type: str):
        if provider_type == "s3":
            if self.default_s3_provider is None:
                self.default_s3_provider = S3Provider("s3://bucket/path")
            return self.default_s3_provider
        else:
            if self.default_gcs_provider is None:
                from hub.core.storage.gcs import GCSProvider

                self.default_gcs_provider = GCSProvider("gcs://bucket/path")
            return self.default_gcs_provider

    def get_storage_provider(self, key: Optional[str], provider_type: str):
        assert provider_type in {"s3", "gcs"}
        if key in {"ENV", None}:
            return self.get_default_provider(provider_type)

        provider: StorageProvider
        creds = self.get_creds(key)

        if provider_type == "s3":
            if key in self.storage_providers:
                provider = self.storage_providers[key]
                if isinstance(provider, S3Provider):
                    return provider

            provider = S3Provider("s3://bucket/path", **creds)
        else:
            from hub.core.storage.gcs import GCSProvider

            if key in self.storage_providers:
                provider = self.storage_providers[key]
                if isinstance(provider, GCSProvider):
                    return provider

            provider = GCSProvider("gcs://bucket/path", **creds)
        self.storage_providers[key] = provider
        return provider

    def add_creds_key(self, creds_key: str, managed: bool = False):
        if creds_key in self.creds_keys:
            raise ValueError(f"Creds key {creds_key} already exists")
        if managed:
            creds = self.fetch_managed_creds(creds_key)
        self.creds_keys.append(creds_key)
        self.creds_mapping[creds_key] = len(self.creds_keys)
        if managed:
            self.managed_creds_keys.add(creds_key)
            self.populate_creds(creds_key, creds)

    def replace_creds(self, old_creds_key: str, new_creds_key: str):
        if old_creds_key not in self.creds_keys:
            raise KeyError(f"Creds key {old_creds_key} does not exist")
        if new_creds_key in self.creds_keys:
            raise ValueError(f"Creds key {new_creds_key} already exists")
        for i in range(len(self.creds_keys)):
            if self.creds_keys[i] == old_creds_key:
                self.creds_keys[i] = new_creds_key
                replaced_index = i

        if old_creds_key in self.creds_dict:
            self.creds_dict[new_creds_key] = self.creds_dict[old_creds_key]
            del self.creds_dict[old_creds_key]

        self.creds_mapping[new_creds_key] = self.creds_mapping[old_creds_key]
        del self.creds_mapping[old_creds_key]

        if old_creds_key in self.managed_creds_keys:
            self.managed_creds_keys.remove(old_creds_key)
            self.managed_creds_keys.add(new_creds_key)

        if old_creds_key in self.used_creds_keys:
            self.used_creds_keys.remove(old_creds_key)
            self.used_creds_keys.add(new_creds_key)

        if old_creds_key in self.storage_providers:
            self.storage_providers[new_creds_key] = self.storage_providers[
                old_creds_key
            ]
            del self.storage_providers[old_creds_key]
        return replaced_index

    def populate_creds(self, creds_key: str, creds):
        if creds_key not in self.creds_keys:
            raise KeyError(f"Creds key {creds_key} does not exist")
        expires_in_to_expires_at(creds)
        self.creds_dict[creds_key] = creds

    def add_to_used_creds(self, creds_key: Optional[str]):
        if creds_key is None or creds_key in self.used_creds_keys:
            return False
        self.used_creds_keys.add(creds_key)
        return True

    def tobytes(self) -> bytes:
        d = {
            "creds_keys": self.creds_keys,
            "managed_creds_keys": list(self.managed_creds_keys),
            "used_creds_keys": list(self.used_creds_keys),
        }
        return json.dumps(d).encode("utf-8")

    @classmethod
    def frombuffer(cls, buffer: bytes):
        obj = cls()
        if buffer:
            d = json.loads(buffer.decode("utf-8"))
            obj.creds_keys = list(d["creds_keys"])
            obj.creds_mapping = {k: i + 1 for i, k in enumerate(obj.creds_keys)}
            obj.managed_creds_keys = set(d["managed_creds_keys"])
            obj.used_creds_keys = set(d["used_creds_keys"])
        obj.is_dirty = False
        return obj

    def get_encoding(self, key: Optional[str] = None, path: Optional[str] = None):
        if key == "ENV":
            return 0
        if key is None:
            if path and path.startswith(ALL_CLOUD_PREFIXES):
                raise ValueError("Creds key must always be specified for cloud storage")
            return 0

        if key not in self.creds_keys:
            raise ValueError(f"Creds key {key} does not exist")
        return self.creds_mapping[key]

    def get_creds_key(self, encoding):
        if encoding > len(self.creds_keys):
            raise KeyError(f"Encoding {encoding} not found.")
        return None if encoding == 0 else self.creds_keys[encoding - 1]

    @property
    def nbytes(self):
        return len(self.tobytes())

    def __getstate__(self):
        return {
            "creds_keys": self.creds_keys,
            "creds_dict": self.creds_dict,
            "managed_creds_keys": self.managed_creds_keys,
            "used_creds_keys": self.used_creds_keys,
        }

    def __setstate__(self, state):
        self.creds_keys = state["creds_keys"]
        self.creds_dict = state["creds_dict"]
        self.managed_creds_keys = state["managed_creds_keys"]
        self.used_creds_keys = state["used_creds_keys"]
        self.creds_mapping = {key: i + 1 for i, key in enumerate(self.creds_keys)}
        self.storage_providers = {}
        self.default_s3_provider = None
        self.default_gcs_provider = None
        self.client = None
        self.org_id = None

    def __len__(self):
        return len(self.creds_keys)

    @property
    def missing_keys(self) -> list:
        return [key for key in self.creds_keys if key not in self.creds_dict]

    def populate_all_managed_creds(self):
        assert self.client is not None
        assert self.org_id is not None
        for creds_key in self.managed_creds_keys:
            creds = self.fetch_managed_creds(creds_key)
            self.populate_creds(creds_key, creds)

    def fetch_managed_creds(self, creds_key: str):
        creds = self.client.get_managed_creds(self.org_id, creds_key)
        print(f"Loaded credentials '{creds_key}' from Activeloop platform.")
        return creds

    def change_creds_management(self, creds_key: str, managed: bool) -> bool:
        if creds_key not in self.creds_keys:
            raise KeyError(f"Creds key {creds_key} not found.")
        is_managed = creds_key in self.managed_creds_keys
        if is_managed == managed:
            return False
        if managed:
            creds = self.fetch_managed_creds(creds_key)
            self.managed_creds_keys.add(creds_key)
            self.populate_creds(creds_key, creds)
        else:
            self.managed_creds_keys.discard(creds_key)

        return True

    def warn_missing_managed_creds(self):
        """Warns about any missing managed creds that were added in parallel by someone else."""
        missing_creds = self.missing_keys

        missing_managed_creds = [
            creds for creds in missing_creds if creds in self.managed_creds_keys
        ]
        if missing_managed_creds:
            warnings.warn(
                f"There are some managed creds missing ({missing_managed_creds}) that were added after the dataset was loaded. Reload the dataset to load them."
            )



from typing import Optional
from hub.constants import ALL_CLOUD_PREFIXES


def convert_creds_key(creds_key: Optional[str], path: str):
    if creds_key is None and path.startswith(ALL_CLOUD_PREFIXES):
        creds_key = "ENV"
    elif creds_key == "ENV" and not path.startswith(ALL_CLOUD_PREFIXES):
        creds_key = None
    return creds_key


class LinkedSample:
    def __init__(self, path: str, creds_key: Optional[str] = None):
        self.path = path
        self.creds_key = convert_creds_key(creds_key, path)



import numpy as np
import pytest
from hub.util.exceptions import EmptyTensorError

compressions_paremetrized = pytest.mark.parametrize(
    "compression",
    [
        {"sample_compression": None},
        {"sample_compression": "png"},
        {"chunk_compression": "png"},
        {"sample_compression": "lz4"},
        {"chunk_compression": "lz4"},
    ],
)


@compressions_paremetrized
@pytest.mark.parametrize("create_shape_tensor", [True, False])
@pytest.mark.parametrize("empty_sample", [None, []])
def test_none_append(local_ds, compression, create_shape_tensor, empty_sample):
    with local_ds as ds:
        ds.create_tensor("xyz", create_shape_tensor=create_shape_tensor, **compression)
        ds.xyz.append(empty_sample)
        ds.xyz.append(empty_sample)
        ds.xyz.append(np.ones((100, 100, 3), dtype=np.uint8))

        for i in range(2):
            assert ds.xyz[i].numpy().shape == (0, 0, 0)
            assert ds.xyz[i].shape == (0, 0, 0)
        assert ds.xyz[2].numpy().shape == (100, 100, 3)
        assert ds.xyz[2].shape == (100, 100, 3)


@compressions_paremetrized
@pytest.mark.parametrize("create_shape_tensor", [True, False])
@pytest.mark.parametrize("empty_sample", [None, []])
def test_only_nones_append(local_ds, compression, create_shape_tensor, empty_sample):
    with local_ds as ds:
        ds.create_tensor("xyz", create_shape_tensor=create_shape_tensor, **compression)
        ds.xyz.append(empty_sample)
        ds.xyz.append(empty_sample)

        for i in range(2):
            with pytest.raises(EmptyTensorError):
                ds.xyz[i].numpy()
            assert ds.xyz[i].shape == ()


@compressions_paremetrized
@pytest.mark.parametrize("create_shape_tensor", [True, False])
@pytest.mark.parametrize("empty_sample", [None, []])
def test_none_updates(local_ds, compression, create_shape_tensor, empty_sample):
    with local_ds as ds:
        ds.create_tensor("xyz", create_shape_tensor=create_shape_tensor, **compression)
        ds.xyz.append(np.ones((100, 100, 3), dtype=np.uint8))
        ds.xyz.append(np.ones((300, 500, 3), dtype=np.uint8))
        ds.xyz.append(np.ones((300, 500, 3), dtype=np.uint8))

        ds.xyz[1] = empty_sample
        assert ds.xyz[0].numpy().shape == (100, 100, 3)
        assert ds.xyz[0].shape == (100, 100, 3)
        assert ds.xyz[1].numpy().shape == (0, 0, 0)
        assert ds.xyz[1].shape == (0, 0, 0)
        assert ds.xyz[2].numpy().shape == (300, 500, 3)
        assert ds.xyz[2].shape == (300, 500, 3)


@pytest.mark.parametrize("empty_sample", [None, []])
def test_none_image_chunk_compression_2d(local_ds, empty_sample):
    with local_ds as ds:
        ds.create_tensor("xyz", chunk_compression="png")
        ds.xyz.append(empty_sample)
        ds.xyz.append(empty_sample)
        assert ds.xyz.meta.max_shape == [0, 0, 0]
        assert ds.xyz[0].shape == ()
        assert ds.xyz[1].shape == ()
        ds.xyz.append(np.ones((500, 500), "uint8"))
        assert ds.xyz.meta.max_shape == [500, 500]
        assert ds.xyz[0].numpy().shape == (0, 0)
        assert ds.xyz[0].shape == (0, 0)
        assert ds.xyz[1].numpy().shape == (0, 0)
        assert ds.xyz[1].shape == (0, 0)
        assert ds.xyz[2].numpy().shape == (500, 500)
        assert ds.xyz[2].shape == (500, 500)


@pytest.mark.parametrize("empty_sample", [None, []])
def test_none_text(local_ds, empty_sample):
    with local_ds as ds:
        ds.create_tensor("xyz", htype="text")
        ds.xyz.append(empty_sample)
        ds.xyz.append(empty_sample)
        assert ds.xyz.meta.max_shape == [1]
        assert ds.xyz[0].numpy().shape == (1,)
        assert ds.xyz[0].shape == (1,)
        assert ds.xyz[0].numpy() == ""
        assert ds.xyz[1].numpy().shape == (1,)
        assert ds.xyz[1].shape == (1,)
        assert ds.xyz[1].numpy() == ""

        ds.xyz.append("hello")
        assert ds.xyz.meta.max_shape == [1]
        assert ds.xyz[0].numpy().shape == (1,)
        assert ds.xyz[0].shape == (1,)
        assert ds.xyz[0].numpy() == ""
        assert ds.xyz[1].numpy().shape == (1,)
        assert ds.xyz[1].shape == (1,)
        assert ds.xyz[1].numpy() == ""
        assert ds.xyz[2].numpy().shape == (1,)
        assert ds.xyz[2].shape == (1,)
        assert ds.xyz[2].numpy() == "hello"


@pytest.mark.parametrize("empty_sample", [None, []])
def test_none_json(local_ds, empty_sample):
    with local_ds as ds:
        ds.create_tensor("xyz", htype="json")
        ds.xyz.append(empty_sample)
        ds.xyz.append(empty_sample)
        assert ds.xyz.meta.max_shape == [1]
        assert ds.xyz[0].numpy().shape == (1,)
        assert ds.xyz[0].shape == (1,)
        assert ds.xyz[0].numpy() == {}
        assert ds.xyz[1].numpy().shape == (1,)
        assert ds.xyz[1].shape == (1,)
        assert ds.xyz[1].numpy() == {}

        ds.xyz.append({"hello": "world"})
        assert ds.xyz.meta.max_shape == [1]
        assert ds.xyz[0].numpy().shape == (1,)
        assert ds.xyz[0].shape == (1,)
        assert ds.xyz[0].numpy() == {}
        assert ds.xyz[1].numpy().shape == (1,)
        assert ds.xyz[1].shape == (1,)
        assert ds.xyz[1].numpy() == {}
        assert ds.xyz[2].numpy().shape == (1,)
        assert ds.xyz[2].shape == (1,)
        assert ds.xyz[2].numpy() == {"hello": "world"}


@pytest.mark.parametrize("empty_sample", [None, []])
def test_none_list(local_ds, empty_sample):
    with local_ds as ds:
        ds.create_tensor("xyz", htype="list")
        ds.xyz.append(empty_sample)
        ds.xyz.append(empty_sample)
        assert ds.xyz.meta.max_shape == [0]
        assert ds.xyz[0].numpy().shape == (0,)
        assert ds.xyz[0].shape == (0,)
        assert ds.xyz[0].numpy().tolist() == []
        assert ds.xyz[1].numpy().shape == (0,)
        assert ds.xyz[1].shape == (0,)
        assert ds.xyz[1].numpy().tolist() == []

        ds.xyz.append(["hello", "world"])
        assert ds.xyz.meta.max_shape == [2]
        assert ds.xyz[0].numpy().shape == (0,)
        assert ds.xyz[0].shape == (0,)
        assert ds.xyz[0].numpy().tolist() == []
        assert ds.xyz[1].numpy().shape == (0,)
        assert ds.xyz[1].shape == (0,)
        assert ds.xyz[1].numpy().tolist() == []
        assert ds.xyz[2].numpy().shape == (2,)
        assert ds.xyz[2].shape == (2,)
        assert ds.xyz[2].numpy().tolist() == ["hello", "world"]



from typing import List
from hub.util.exceptions import AgreementNotAcceptedError


def agreement_prompt(agreements: List[str], org_id: str, ds_name: str):
    print(
        "\n\nThe owner of the dataset you are trying to access requires that you agree to the following terms first:\n"
    )
    print("-" * 16)
    print(f"Dataset Owner: {org_id}")
    print(f"Dataset Name: {ds_name}")
    print("-" * 16)

    print("Terms:")

    for agreement in agreements:
        print(agreement + "\n")

    print("-" * 16)
    user_input = input(
        f"In order to accept, please type the dataset's name ({ds_name}) and press enter: "
    )
    return user_input == ds_name


def handle_dataset_agreements(client, agreements: List[str], org_id: str, ds_name: str):
    accepted = agreement_prompt(agreements, org_id, ds_name)
    if not accepted:
        raise AgreementNotAcceptedError()
    client.accept_agreements(org_id, ds_name)
    print("Accepted agreement!")



from hub.util.exceptions import InvalidHubPathException
from typing import Tuple
import hub


def process_hub_path(path: str) -> Tuple[str, str, str, str]:
    """Checks whether path is a valid hub path."""
    # Allowed formats:
    # hub://org/ds
    # hub://org/ds/.queries/hash
    # hub://org/queries/hash
    # hub://org/queries/.queries/hash
    # hub://org/ds/sub_ds1/sub_ds2/sub_ds3/..../sub_ds{n}  # Only for internal usage.

    tag = path[6:]
    s = tag.split("/")

    if len(s) < 2:
        raise InvalidHubPathException(path)

    path = f"hub://{s[0]}/{s[1]}"

    if len(s) == 3 and s[1] == "queries" and not s[2].startswith("."):
        # Special case: expand hub://username/queries/hash to hub://username/queries/.queries/hash
        subdir = f".queries/{s[2]}"
    else:
        subdir = "/".join(s[2:])
        if len(s) > 2:
            if (
                not (len(s) == 4 and s[2] == ".queries")
                and not hub.constants._ENABLE_HUB_SUB_DATASETS
            ):
                raise InvalidHubPathException(path)
    return (path, *s[:2], subdir)  # type: ignore



from multiprocessing import Manager
from hub.core.compute.provider import ComputeProvider
from pathos.pools import ThreadPool  # type: ignore


class ThreadProvider(ComputeProvider):
    def __init__(self, workers):
        self.workers = workers
        self.manager = Manager()
        self.pool = ThreadPool(nodes=workers)

    def map(self, func, iterable):
        return self.pool.map(func, iterable)

    def create_queue(self):
        return self.manager.Queue()

    def close(self):
        self.pool.close()
        self.pool.join()
        self.pool.clear()



import hub
import pytest
import numpy as np


@pytest.mark.parametrize(
    "kwargs",
    [
        {"sample_shape": (1003, 1103, 3), "tile_shape": (10, 10, 3)},
        {"sample_shape": (100003, 300007, 3)},
    ],
)
@pytest.mark.parametrize(
    "compression_type",
    [
        "sample_compression",
        "chunk_compression",
    ],
)
@pytest.mark.parametrize(
    "compression",
    [
        "png",
        "lz4",
    ],
)
def test_partial_upload(memory_ds, kwargs, compression_type, compression):
    ds = memory_ds
    ds.create_tensor("image", htype="image", **{compression_type: compression})
    ds.image.append(hub.tiled(**kwargs))
    np.testing.assert_array_equal(
        ds.image[0][:10, :10].numpy(), np.zeros((10, 10, 3), dtype=np.uint8)
    )
    r = np.random.randint(0, 256, (217, 212, 2), dtype=np.uint8)
    ds.image[0][-217:, :212, 1:] = r
    np.testing.assert_array_equal(ds.image[0][-217:, :212, 1:].numpy(), r)



from typing import Any, Callable, List, Union, Optional
from hub.core.dataset import Dataset
from hub.core.io import IOBlock, SampleStreaming
from hub.core.index import Index
from hub.core.tensor import Tensor


import numpy as np


NP_RESULT = Union[np.ndarray, List[np.ndarray]]
NP_ACCESS = Callable[[str], NP_RESULT]


class DatasetQuery:
    def __init__(
        self,
        dataset,
        query: str,
        progress_callback: Callable[[int, bool], None] = lambda *_: None,
    ):
        self._dataset = dataset
        self._query = query
        self._pg_callback = progress_callback
        self._cquery = compile(query, "", "eval")
        self._tensors = [
            tensor
            for tensor in dataset.tensors.keys()
            if normalize_query_tensors(tensor) in query
        ]
        self._blocks = expand(dataset, self._tensors)
        self._np_access: List[NP_ACCESS] = [
            _get_np(dataset, block) for block in self._blocks
        ]
        self._wrappers = self._export_tensors()
        self._groups = self._export_groups(self._wrappers)

    def execute(self) -> List[int]:
        idx_map: List[int] = list()

        for f, blk in zip(self._np_access, self._blocks):
            cache = {tensor: f(tensor) for tensor in self._tensors}
            for local_idx in range(len(blk)):
                p = {
                    tensor: self._wrap_value(tensor, cache[tensor][local_idx])
                    for tensor in self._tensors
                }
                p.update(self._groups)
                if eval(self._cquery, p):
                    global_index = blk.indices()[local_idx]
                    idx_map.append(global_index)
                    self._pg_callback(local_idx, True)
                else:
                    self._pg_callback(local_idx, False)
        return idx_map

    def _wrap_value(self, tensor, val):
        if tensor in self._wrappers:
            return self._wrappers[tensor].with_value(val)
        else:
            return val

    def _export_tensors(self):
        return {
            tensor_key: export_tensor(tensor)
            for tensor_key, tensor in self._dataset.tensors.items()
        }

    def _export_groups(self, wrappers):
        return {
            extract_prefix(tensor_key): GroupTensor(
                self._dataset, wrappers, extract_prefix(tensor_key)
            )
            for tensor_key in self._dataset.tensors.keys()
            if "/" in tensor_key
        }


def normalize_query_tensors(tensor_key: str) -> str:
    return tensor_key.replace("/", ".")


def extract_prefix(tensor_key: str) -> str:
    return tensor_key.split("/")[0]


def _get_np(dataset: Dataset, block: IOBlock) -> NP_ACCESS:
    idx = block.indices()

    def f(tensor: str) -> NP_RESULT:
        tensor_obj = dataset.tensors[tensor]
        tensor_obj.index = Index()
        return tensor_obj[idx]

    return f


def expand(dataset, tensor: List[str]) -> List[IOBlock]:
    return SampleStreaming(dataset, tensor).list_blocks()


def export_tensor(tensor: Tensor):
    if tensor.htype == "class_label":
        return ClassLabelsTensor(tensor)

    return EvalObject()


class EvalObject:
    def __init__(self) -> None:
        self._val: Any = None
        self._numpy: Optional[Union[np.ndarray, List[np.ndarray]]] = None

    @property
    def value(self):
        return self._val

    def with_value(self, v: Any):
        self._val = v
        self._numpy = None
        return self

    @property
    def numpy_value(self):
        if self._numpy is None:
            self._numpy = self._val.numpy(
                aslist=self._val.is_dynamic, fetch_chunks=True
            )
        return self._numpy

    def contains(self, v: Any):
        return v in self.numpy_value

    def __getitem__(self, item):
        r = EvalObject()
        return r.with_value(self.value[item])

    @property
    def min(self):
        """Returns np.min() for the tensor"""
        return np.amin(self.numpy_value)

    @property
    def max(self):
        """Returns np.max() for the tensor"""
        return np.amax(self.numpy_value)

    @property
    def mean(self):
        """Returns np.mean() for the tensor"""
        return self.numpy_value.mean()

    @property
    def shape(self):
        """Returns shape of the underlying numpy array"""
        return self.value.shape  # type: ignore

    @property
    def size(self):
        """Returns size of the underlying numpy array"""
        return self.value.size  # type: ignore

    def _to_np(self, o):
        if isinstance(o, EvalObject):
            return o.numpy_value
        return o

    def __eq__(self, o: object) -> bool:
        val = self.numpy_value
        o = self._to_np(o)
        if isinstance(val, (list, np.ndarray)):
            if isinstance(o, (list, tuple)):
                return set(o) == set(val)
            else:
                return o in val
        else:
            return val == o

    def __lt__(self, o: object) -> bool:
        o = self._to_np(o)
        return self.numpy_value < o

    def __le__(self, o: object) -> bool:
        o = self._to_np(o)
        return self.numpy_value <= o

    def __gt__(self, o: object) -> bool:
        o = self._to_np(o)
        return self.numpy_value > o

    def __ge__(self, o: object) -> bool:
        o = self._to_np(o)
        return self.numpy_value >= o

    def __mod__(self, o: object):
        o = self._to_np(o)
        return self.numpy_value % o

    def __add__(self, o: object):
        o = self._to_np(o)
        return self.numpy_value + o

    def __sub__(self, o: object):
        o = self._to_np(o)
        return self.numpy_value - o

    def __div__(self, o: object):
        o = self._to_np(o)
        return self.numpy_value / o

    def __floordiv__(self, o: object):
        o = self._to_np(o)
        return self.numpy_value // o

    def __mul__(self, o: object):
        o = self._to_np(o)
        return self.numpy_value * o

    def __pow__(self, o: object):
        o = self._to_np(o)
        return self.numpy_value**o

    def __contains__(self, o: object):
        o = self._to_np(o)
        return self.contains(o)

    @property
    def sample_info(self):
        return self._val.sample_info


class GroupTensor:
    def __init__(self, dataset: Dataset, wrappers, prefix: str) -> None:
        self.prefix = prefix
        self.dataset = dataset
        self.wrappers = wrappers
        self._subgroup = self.expand()

    def __getattr__(self, __name: str) -> Any:
        return self._subgroup[self.normalize_key(__name)]

    def expand(self):
        r = {}
        for tensor in [
            self.normalize_key(t)
            for t in self.dataset.tensors
            if t.startswith(self.prefix)
        ]:
            prefix = self.prefix + "/" + extract_prefix(tensor)
            if "/" in tensor:
                r[tensor] = GroupTensor(self.dataset, self.wrappers, prefix)
            else:
                r[tensor] = self.wrappers[prefix]

        return r

    def normalize_key(self, key: str) -> str:
        return key.replace(self.prefix + "/", "")


class ClassLabelsTensor(EvalObject):
    def __init__(self, tensor: Tensor) -> None:
        super(ClassLabelsTensor, self).__init__()
        _classes = tensor.info["class_names"]  # type: ignore
        self._classes_dict = {v: idx for idx, v in enumerate(_classes)}

    def _norm_labels(self, o: object):
        o = self._to_np(o)
        if isinstance(o, str):
            return self._classes_dict[o]
        elif isinstance(o, int):
            return o
        elif isinstance(o, (list, tuple)):
            return o.__class__(map(self._norm_labels, o))

    def __eq__(self, o: object) -> bool:
        try:
            o = self._norm_labels(o)
        except KeyError:
            return False
        return super(ClassLabelsTensor, self).__eq__(o)

    def __lt__(self, o: object) -> bool:
        o = self._to_np(o)
        if isinstance(o, str):
            raise ValueError("label class is not comparable")
        return self.numpy_value < o

    def __le__(self, o: object) -> bool:
        o = self._to_np(o)
        if isinstance(o, str):
            raise ValueError("label class is not comparable")
        return self.numpy_value <= o

    def __gt__(self, o: object) -> bool:
        o = self._to_np(o)
        if isinstance(o, str):
            raise ValueError("label class is not comparable")
        return self.numpy_value > o

    def __ge__(self, o: object) -> bool:
        o = self._to_np(o)
        if isinstance(o, str):
            raise ValueError("label class is not comparable")
        return self.numpy_value >= o

    def contains(self, v: Any):
        v = self._to_np(v)
        if isinstance(v, str):
            v = self._classes_dict[v]
        return super(ClassLabelsTensor, self).contains(v)



from hub.core.transform.transform_tensor import TransformTensor
from hub.util.exceptions import TensorDoesNotExistError


class TransformDataset:
    def __init__(self, all_tensors=None, slice_list=None):
        """Creates a Dataset like object that supports "." access of tensors and appends/extends to the tensors.
        This is used as sample_out in hub transforms.
        """
        self.tensors = all_tensors or {}
        self.slice_list = slice_list or []

    def __len__(self):
        return min(len(self[tensor]) for tensor in self.tensors)

    def __getattr__(self, name):
        if name not in self.tensors:
            self.tensors[name] = TransformTensor(name=name, dataset=self)
        return self.tensors[name][self.slice_list]

    def __getitem__(self, slice_):
        if isinstance(slice_, str):
            return self.__getattr__(slice_)
        assert isinstance(slice_, (slice, int))
        new_slice_list = self.slice_list + [slice_]
        return TransformDataset(all_tensors=self.tensors, slice_list=new_slice_list)

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]

    def append(self, sample, skip_ok=False):
        if not skip_ok:
            for k in self.tensors:
                if k not in sample:
                    raise TensorDoesNotExistError(k)
        if len(set(map(len, (self[k] for k in sample)))) != 1:
            raise ValueError("All tensors are expected to have the same length.")
        for k, v in sample.items():
            self[k].append(v)



import numpy as np
from typing import List, Sequence, Union
import warnings

from hub.core.chunk.base_chunk import InputSample
from hub.core.index.index import Index
from hub.util.exceptions import DynamicTensorNumpyError


# used for warning the user if updating a tensor caused suboptimal chunks
CHUNK_UPDATE_WARN_PORTION = 0.2


def format_read_samples(
    samples: List[np.ndarray], index: Index, aslist: bool
) -> Union[np.ndarray, List[np.ndarray]]:
    """Prepares samples being read from the chunk engine in the format the user expects."""

    samples = index.apply(samples)  # type: ignore

    if aslist and all(map(np.isscalar, samples)):
        samples = [arr.item() for arr in samples]

    samples = index.apply_squeeze(samples)  # type: ignore

    if aslist:
        return samples
    else:
        return np.array(samples)


def check_samples_type(samples):
    if not isinstance(samples, (List, np.ndarray)):
        raise TypeError(f"Cannot extend with samples of type {type(samples)}")


def make_sequence(
    samples: Union[np.ndarray, Sequence[InputSample], InputSample], index_length: int
) -> Sequence[InputSample]:
    """Make `samples` a sequence of `InputSample`s.

    Args:
        samples (Union[np.ndarray, Sequence[InputSample]]): Incoming samples to be made into a sequence.
        index_length (int): Number of expected samples in the sequence.

    Raises:
        ValueError: If `index_length` is incompatible with the true length of `samples`.

    Returns:
        Sequence[InputSample]: Sequence of `InputSample`s with the same length as `index_length`.
    """

    if index_length == 1:
        if hasattr(samples, "__len__"):
            if len(samples) != 1:  # type: ignore
                samples = [samples]
        elif hasattr(samples, "shape"):
            if len(samples.shape) > 0 and samples.shape[0] != 1:  # type: ignore
                samples = [samples]
        else:
            samples = [samples]

    if hasattr(samples, "__len__"):
        if index_length != len(samples):  # type: ignore
            raise ValueError(
                f"Index length ({index_length}) and length of samples ({len(samples)}) must be equal for updating a tensor."  # type: ignore
            )
    else:
        samples = [samples]

    return samples  # type: ignore


def check_suboptimal_chunks(
    chunks_nbytes_after_updates: List[int], min_chunk_size: int, max_chunk_size: int
):
    upper_warn_threshold = max_chunk_size * (1 + CHUNK_UPDATE_WARN_PORTION)
    lower_warn_threshold = min_chunk_size * (1 - CHUNK_UPDATE_WARN_PORTION)

    for nbytes in chunks_nbytes_after_updates:
        if nbytes > upper_warn_threshold or nbytes < lower_warn_threshold:
            warnings.warn(
                "After update, some chunks were suboptimal. Be careful when doing lots of updates that modify the sizes of samples by a large amount, these can heavily impact read performance!"
            )
            break


def check_sample_shape(shape, last_shape, key, index, aslist):
    if not aslist and last_shape is not None and shape != last_shape:
        raise DynamicTensorNumpyError(key, index, "shape")



import pytest


def test_dataset(local_ds_generator):
    ds = local_ds_generator()

    assert len(ds.info) == 0

    ds.info.update(my_key=0)
    ds.info.update(my_key=1)

    ds.info.update(another_key="hi")
    ds.info.update({"another_key": "hello"})

    ds.info.update({"something": "aaaaa"}, something="bbbb")

    ds.info.test = [1, 2, "5"]

    test_list = ds.info.test
    with ds:
        ds.info.update({"test2": (1, 5, (1, "2"), [5, 6, (7, 8)])})
        ds.info.update(xyz="abc")
        test_list.extend(["user made change without `update`"])

    ds.info.update({"1_-+": 5})
    assert len(ds.info) == 7

    ds = local_ds_generator()

    assert len(ds.info) == 7

    assert ds.info.another_key == "hello"
    assert ds.info.something == "bbbb"

    assert ds.info.test == [1, 2, "5", "user made change without `update`"]
    assert ds.info.test2 == [1, 5, [1, "2"], [5, 6, [7, 8]]]

    assert ds.info.xyz == "abc"
    assert ds.info["1_-+"] == 5  # key can't be accessed with `.` syntax

    ds.info.update(test=[99])

    ds = local_ds_generator()

    assert len(ds.info) == 7
    assert ds.info.test == [99]

    ds.info.pop("test")
    assert len(ds.info) == 6

    ds.info.pop("1_-+")
    ds.info.pop("xyz")
    assert len(ds.info) == 4

    ds.info.clear()
    assert len(ds.info) == 0


def test_tensor(local_ds_generator):
    ds = local_ds_generator()

    t1 = ds.create_tensor("tensor1")
    t2 = ds.create_tensor("tensor2")

    t1.info.update(key=0)
    t2.info.update(key=1, key1=0)
    t2.info.key2 = 2
    t2.info["key3"] = 3

    ds = local_ds_generator()

    t1 = ds.tensor1
    t2 = ds.tensor2

    assert len(t1.info) == 1
    assert len(t2.info) == 4

    assert t1.info.key == t1.info["key"] == 0
    assert t2.info.key == t2.info["key"] == 1
    assert t2.info.key1 == t2.info["key1"] == 0
    assert t2.info.key2 == t2.info["key2"] == 2
    assert t2.info.key3 == t2.info["key3"] == 3

    with ds:
        t1.info.update(key=99)

    ds = local_ds_generator()

    t1 = ds.tensor1
    t2 = ds.tensor2

    assert len(t1.info) == 1
    assert len(t2.info) == 4

    assert t1.info.key == 99

    t2.info.pop("key")
    assert len(t2.info) == 3

    t2.info.pop("key2")
    t2.info.pop("key3")
    assert len(t2.info) == 1

    t2.info.clear()
    assert len(t2.info) == 0


def test_update_reference_manually(local_ds_generator):
    """Right now synchronization can only happen when you call `info.update`."""

    ds = local_ds_generator()

    ds.info.update(key=[1, 2, 3])

    ds = local_ds_generator()

    l = ds.info.key
    assert l == [1, 2, 3]

    # un-registered update
    l.append(5)
    assert ds.info.key == [1, 2, 3, 5]

    ds = local_ds_generator()

    l = ds.info.key
    assert l == [1, 2, 3]

    # registered update
    l.append(99)
    ds.info.update()

    ds = local_ds_generator()

    assert l == [1, 2, 3, 99]


def test_class_label(local_ds_generator):
    ds = local_ds_generator()
    ds.create_tensor("labels", htype="class_label", class_names=["a", "b", "c"])
    ds.create_tensor("labels2", htype="class_label")
    assert len(ds.labels.info) == 1
    assert len(ds.labels2.info) == 1
    assert (
        ds.labels.info.class_names == ds.labels.info["class_names"] == ["a", "b", "c"]
    )
    assert ds.labels2.info.class_names == ds.labels2.info["class_names"] == []
    ds.labels.info.class_names = ["c", "b", "a"]
    ds = local_ds_generator()
    assert len(ds.labels.info) == 1
    assert len(ds.labels2.info) == 1
    assert (
        ds.labels.info.class_names == ds.labels.info["class_names"] == ["c", "b", "a"]
    )
    assert ds.labels2.info.class_names == ds.labels2.info["class_names"] == []


def test_bbox(local_ds_generator):
    ds = local_ds_generator()
    ds.create_tensor("bboxes", htype="bbox", coords={"type": 0, "mode": 2})
    ds.create_tensor("bboxes1", htype="bbox", coords={"type": 1})
    ds.create_tensor("bboxes2", htype="bbox")
    assert len(ds.bboxes.info) == 1
    assert len(ds.bboxes2.info) == 1
    assert ds.bboxes.info.coords == ds.bboxes.info["coords"] == {"type": 0, "mode": 2}
    assert ds.bboxes1.info.coords == ds.bboxes1.info["coords"] == {"type": 1}
    assert ds.bboxes2.info.coords == ds.bboxes2.info["coords"] == {}
    ds.bboxes.info.coords = {"type": 3}
    ds = local_ds_generator()
    assert len(ds.bboxes.info) == 1
    assert len(ds.bboxes2.info) == 1
    assert ds.bboxes.info.coords == ds.bboxes.info["coords"] == {"type": 3}
    assert ds.bboxes2.info.coords == ds.bboxes2.info["coords"] == {}
    with pytest.raises(TypeError):
        ds.create_tensor("bboxes3", htype="bbox", coords=[1, 2, 3])

    with pytest.raises(KeyError):
        ds.create_tensor("bboxes4", htype="bbox", coords={"random": 0})


def test_info_new_methods(local_ds_generator):
    ds = local_ds_generator()
    ds.create_tensor("x")

    ds.info[0] = "hello"
    ds.info[1] = "world"
    assert len(ds.info) == 2
    assert set(ds.info.keys()) == {0, 1}
    assert 0 in ds.info
    assert 1 in ds.info

    assert ds.info[0] == "hello"
    assert ds.info[1] == "world"

    del ds.info[0]
    assert len(ds.info) == 1
    assert 1 in ds.info
    assert ds.info[1] == "world"

    for it in ds.info:
        assert it == 1

    ds.info.setdefault(0, "yo")
    assert len(ds.info) == 2
    assert 0 in ds.info
    assert 1 in ds.info
    assert ds.info[0] == "yo"
    assert ds.info[1] == "world"

    ds.info.popitem()
    assert len(ds.info) == 1
    assert 1 in ds.info
    assert ds.info[1] == "world"

    for k, v in ds.info.items():
        assert k == 1
        assert v == "world"

    for v in ds.info.values():
        assert v == "world"

    ds.info = {"a": "b"}
    assert len(ds.info) == 1
    assert "a" in ds.info
    assert ds.info["a"] == "b"

    ds.x.info = {"x": "y", "z": "w"}
    assert len(ds.x.info) == 2
    assert "x" in ds.x.info
    assert "z" in ds.x.info
    assert ds.x.info["x"] == "y"
    assert ds.x.info["z"] == "w"

    with pytest.raises(TypeError):
        ds.info = ["abc"]

    with pytest.raises(TypeError):
        ds.x.info = ["abc"]


def test_info_persistence_bug(local_ds_generator):
    ds = local_ds_generator()
    with ds:
        ds.create_tensor("xyz")
    ds.commit()
    ds.xyz.info.update(abc=123)
    assert ds.xyz.info.abc == 123

    ds = local_ds_generator()
    assert ds.xyz.info.abc == 123


def test_info_has_head_changes_bug(local_ds):
    with local_ds as ds:
        ds.info["tst"] = 42
    assert ds.has_head_changes == True
